#include "stdafx.h"
#include "Archiver.h"
#include "../File.h"
#include "../TempFile.h"
#include "../String/Encoding.h"
namespace Themis {

vector<string> IArchiver::GetSupportedSuffixesLc() const {
    vector<string> vct = InnerGetSupportedSuffixes();
    for (size_t i = 0, n = vct.size(); i < n; i++) {
        vct[i] = vct[i].ToLower();
    }
    return vct;
}

IArchiver::SFile::SFile(const Detail::INDIVIDUALINFO& ii) {
    FILETIME ft = {0};
    BOOL bRet = ::DosDateTimeToFileTime(ii.wDate, ii.wTime, &ft);
    Themis_DebugAssertApi(bRet != FALSE);
    // oփZbgB
    strName = Encoding::GetDefault().GetString(ii.szFileName);
    nOrgSize = ii.dwOriginalSize;
    ftLastUpdate = ft;
}

void IArchiver::Open(const CFileName& oFileName) {
    m_strFileName = oFileName.Get();
    if (!InnerOpen(m_vctFileList)) {
        throw RuntimeException(oFileName.Get() + L" 𓀏o܂ł");
    }
}

bool IArchiver::OpenNoThrow(const CFileName& oFileName) {
    m_strFileName = oFileName.Get();
    return InnerOpen(m_vctFileList);
}

void IArchiver::CheckBadArchive(int nFileIndex) {
    string strDir = CFileNameSystemDir().Get();
    string strFile = CFileNameSystemDir(GetFileName(nFileIndex)).Get();
    if (strFile.GetLength() < strDir.GetLength() ||
        strFile.Substring(0, strDir.GetLength()) != strDir) {
        throw RuntimeException(GetName() + L" ɉt@C܂F" + GetFileName(nFileIndex));
    }
}

void IArchiver::ExtractAll(const CFileName& oDirName, bool showProgress) {
 // Themis_DebugAssert(GetType() == eTypeSolid);
    // sȃt@Ĉe
    for (int i = 0, n = GetFileCount(); i < n; i++) {
        CheckBadArchive(i);
    }
    // 
    InnerExtractAll(oDirName, showProgress);
}

void IArchiver::ExtractLoad(int nFileIndex, vector<uint8_t>& vctData) {
    Themis_DebugLog(L"IArchiver::ExtractLoad()Ă΂ꂽF" + GetName());
    Themis_DebugAssert(GetType() == eTypeNormal || GetType() == eTypeSolid);
    CTempDir oDir;
    oDir.Create();
    InnerExtractFile(nFileIndex, oDir.GetName());
    CFile(CFileName(oDir.GetName().Get() + GetFileName(nFileIndex))).Load(vctData);
}

void IArchiver::LoadExtract(int nFileIndex, const CFileName& oDirName) {
    Themis_DebugLog(L"IArchiver::LoadExtract()Ă΂ꂽF" + GetName());
    Themis_DebugAssert(GetType() == eTypeLoadable);
    vector<uint8_t> vctData;
    InnerLoadFile(nFileIndex, vctData);
    CFile(CFileName(GetFileName(nFileIndex), oDirName)).Save(vctData);
}

/// ɓt@CCfbNX̎擾BȂ-1
int IArchiver::Find(const string& strFileName) const {
    string strTmp = strFileName.Replace(L"\\", L"/").ToLower();
    for (int i = 0, n = GetFileCount(); i < n; i++) {
        if (GetFileName(i).Replace(L"\\", L"/").ToLower() == strTmp) {
            return i; // vۂB
        }
    }
    return -1;
}

//------------------------------------------------------------------//

CArchiver::CArchiver() {
}

CArchiver::~CArchiver() {
}

vector<string> CArchiver::GetSupportedSuffixesLc() {
    vector<string> vct;
    FactoryList<IArchiver>& r = GetFactoryList();
    for (int i = 0, n = r.GetCount(); i < n; i++) {
        try {
            std::tr1::shared_ptr<IArchiver> pArchiver = r.CreateInstance(i);
            vector<string> tmp = pArchiver->GetSupportedSuffixesLc();
            vct.insert(vct.end(), tmp.begin(), tmp.end());
        } catch (std::exception&) {
        }
    }
    if (std::find(vct.begin(), vct.end(), L"zip") != vct.end()) {
        vct.push_back(L"cbz");
    }
    if (std::find(vct.begin(), vct.end(), L"rar") != vct.end()) {
        vct.push_back(L"cbr");
    }
    std::sort(vct.begin(), vct.end());
    vct.erase(std::unique(vct.begin(), vct.end()), vct.end());
    return vct;
}

void CArchiver::Open(const CFileName& oFileName) {
    try {
        string strSuffixLc = oFileName.GetSuffixLc();
        FactoryList<IArchiver>& r = GetFactoryList();
        for (int i = 0, n = r.GetCount(); i < n; i++) {
            std::tr1::shared_ptr<IArchiver> pArchiver = r.CreateInstance(i);
            vector<string> suffixes = pArchiver->GetSupportedSuffixesLc();
            if (std::find(suffixes.begin(), suffixes.end(), oFileName.GetSuffixLc()) != suffixes.end()) {
                if (pArchiver->OpenNoThrow(oFileName)) {
                    m_pArchiver = pArchiver;
                    return;
                }
            }
        }
    } catch (std::exception&) {
        throw;
    }
    m_pArchiver.reset();
    throw RuntimeException(oFileName.Get() + L" ̏ɓǂݍ݂Ɏs܂");
}

FactoryList<IArchiver>& CArchiver::GetFactoryList() {
    static FactoryList<IArchiver> o;
    return o;
}

}
