#include "Scanner.h"
#include "../ErrorListener.h"
#include "../StringScanner.h"
#include "../String/StringBuilder.h"
namespace Themis {
namespace {

/// ʏ̏
class CCalculatorScanner : public ICalculatorScanner {
    std::tr1::shared_ptr<IErrorListener> m_pErrorListener;
    CStringScanner  m_sz;
    int             m_nNum;
    double          m_dNumF;
    StringBuilder m_str;  //eToken_Label, eToken_Str
public:
    CCalculatorScanner(const string& str, const std::tr1::shared_ptr<IErrorListener>& r)
        : m_sz(str), m_pErrorListener(r) {}
private:
    void InnerReset() { m_sz.ResetPos(); }
    EToken InnerToNext() {
        //Xy[XȂǂ͔΂
        m_sz.SkipSpace();
        //g[N̉(pxȂ̂o邾)
        if (!m_sz) {
            return eToken_EOF;
        } else if (m_sz.IsMatch(L'0', L'x')) {
            m_nNum = m_sz.GetHex();
            return eToken_Num;
        } else if (m_sz.IsMatch(L'0', L'b')) {
            m_nNum = m_sz.GetBin();
            return eToken_Num;
        } else if (m_sz.IsDigit()) {
            int n = m_sz.GetDigit();
            if (m_sz.Get1() != L'.') {
                m_nNum = n;
                return eToken_Num;
            } else {
                m_sz.AddPos(-1);
                int t = m_sz.Get1() - L'0';
                m_dNumF = m_sz.GetFloat();
                m_dNumF += n;
                m_dNumF -= t;
                return eToken_NumF;
            }
        } else if (m_sz.IsMatch(L'"')) {
            m_str.Remove(0, m_str.GetLength());
            while (!m_sz.IsMatch(L'"')) {
                if (m_sz.IsMatch(L'\r', L'\n') || m_sz.IsMatch(L'\n')) {
                    if (m_pErrorListener.get() != NULL) {
                        m_pErrorListener->OnError(string() + L"񃊃eɉsR[ho܂");
                    }
                } else if (m_sz.IsMatch(L'\\', L'\\')) {
                    m_str.Append(L'\\');
                } else if (m_sz.IsMatch(L'\\', L'"')) {
                    m_str.Append(L'"');
                } else {
                    m_str.Append(m_sz.Get1());
                    m_sz.AddPos(1);
                }
            }
            return eToken_Str;
        } else if (m_sz.IsMatch(L'+')) { return eToken_Plus;
        } else if (m_sz.IsMatch(L'-')) { return eToken_Minus;
        } else if (m_sz.IsMatch(L'*')) { return eToken_Mul;
        } else if (m_sz.IsMatch(L'/')) { return eToken_Div;
        } else if (m_sz.IsMatch(L'%')) { return eToken_Mod;
        } else if (m_sz.IsMatch(L'(')) { return eToken_LPr;
        } else if (m_sz.IsMatch(L')')) { return eToken_RPr;
        } else if (m_sz.IsMatch(L'[')) { return eToken_LSq;
        } else if (m_sz.IsMatch(L']')) { return eToken_RSq;
        } else if (m_sz.IsMatch(L'^')) { return eToken_Xor;
        } else if (m_sz.IsMatch(L'~')) { return eToken_CPL;
        } else if (m_sz.IsMatch(L',')) { return eToken_Comma;
        } else if (m_sz.IsMatch(L'.')) { return eToken_Dot;
        } else if (m_sz.IsMatch(L'?')) { return eToken_Question;
        } else if (m_sz.IsMatch(L':')) { return eToken_Colon;
        } else if (m_sz.IsMatch(L'&', L'&')) { return eToken_And2;
        } else if (m_sz.IsMatch(L'|', L'|')) { return eToken_Or2;
        } else if (m_sz.IsMatch(L'>', L'>')) { return eToken_ShR;
        } else if (m_sz.IsMatch(L'<', L'<')) { return eToken_ShL;
        } else if (m_sz.IsMatch(L'>', L'=')) { return eToken_Ge;
        } else if (m_sz.IsMatch(L'<', L'=')) { return eToken_Le;
        } else if (m_sz.IsMatch(L'!', L'=')) { return eToken_Ne;
        } else if (m_sz.IsMatch(L'=', L'=')) { return eToken_Eq;
        } else if (m_sz.IsMatch(L'&')) { return eToken_And;
        } else if (m_sz.IsMatch(L'|')) { return eToken_Or;
        } else if (m_sz.IsMatch(L'>')) { return eToken_Gt;
        } else if (m_sz.IsMatch(L'<')) { return eToken_Lt;
        } else if (m_sz.IsMatch(L'!')) { return eToken_Not;
        } else if (m_sz.IsCSymbolFirst() || m_sz.Is2Byte()) {
            m_str.Remove(0, m_str.GetLength());
            do { m_sz.Copy1(m_str); } while (m_sz.IsCSymbol() || m_sz.Is2Byte());
            return eToken_Label;
        } else {
            if (m_pErrorListener.get() != NULL) {
                m_pErrorListener->OnError(L"ȕ '" + string(m_sz.Get1(), 1) + L"' ܂B");
            }
            m_sz.ToNext(); //Ƃ肠΂Ă܂
            return eToken_Error;
        }
    }
    int InnerGetNum() const { return m_nNum; }
    double InnerGetNumF() const { return m_dNumF; }
    string InnerGetStr() const { return m_str.ToString(); }
};

/// ŏɂ܂Ƃ߂ĉ
class CCompiledCalculatorScanner: public ICalculatorScanner {
    struct SToken {
        EToken  m_eToken;
        int     m_nDataIndex;
    };
    union UData {
        int     nNum;
        double  dNumF;
        int     nStrIndex;
    };
    std::vector<SToken>     m_vctToken;
    std::vector<UData>      m_vctData;
    std::vector<string>     m_vctStrData;
    int                     m_nIndex;
    int                     m_nDataIndex;
public:
    CCompiledCalculatorScanner(CCalculatorScanner& r) {
        r.Reset();
        while (true) {
            SToken s;
            s.m_eToken = r.GetToken();
            s.m_nDataIndex = static_cast<int>(m_vctData.size());
            m_vctToken.push_back(s);
            if (s.m_eToken == eToken_EOF) {
                break;
            }
            if (s.m_eToken == eToken_Num) {
                UData u;
                u.nNum = r.GetNum();
                m_vctData.push_back(u);
            } else if (s.m_eToken == eToken_NumF) {
                UData u;
                u.dNumF = r.GetNumF();
                m_vctData.push_back(u);
            } else if (s.m_eToken == eToken_Label || s.m_eToken == eToken_Str) {
                UData u;
                u.nStrIndex = static_cast<int>(m_vctStrData.size());
                m_vctData.push_back(u);
                m_vctStrData.push_back(r.GetStr());
            }
            r.ToNext();
        }
    }
private:
    void InnerReset() { m_nIndex = 0; }
    EToken InnerToNext() {
        m_nDataIndex = m_vctToken[m_nIndex].m_nDataIndex;
        return m_vctToken[m_nIndex++].m_eToken;
    }
    int InnerGetNum() const { return m_vctData[m_nDataIndex].nNum; }
    double InnerGetNumF() const { return m_vctData[m_nDataIndex].dNumF; }
    string InnerGetStr() const { return m_vctStrData[m_vctData[m_nDataIndex].nStrIndex]; }
};

}//namespace
}
