#include "stdafx.h"
#include "FileName.h"
#include "Windows/Environment.h"
#include "String/StringBuilder.h"
#include "String/Encoding.h"
#include "Utility/Algorithm.h"
#include "Utility/StringUtility.h"
namespace Themis {

namespace {
    string GetModuleDir() {
        char sz[_MAX_PATH];
        ::GetModuleFileName(NULL, sz, sizeof(sz));
        //Ō̃fBNg؂T
        string str = Encoding::GetDefault().GetString(sz);
        int n = CFileName::GetLastSeparatorPos(str);
        if (n < 0) {
            throw RuntimeException(L"st@C̃t@CɃfBNg؂蕶 L'\\' ܂");
        }
        return str.Substring(0, n + 1);
    }
    string GetSystemDir() {
        string str = Environment::GetCurrentDirectory();
        return str.EndsWith(L"\\") ? str : str + L"\\";
    }
}

CFileNameModuleDir::CFileNameModuleDir(): CFileName(GetModuleDir()) {
}

CFileNameModuleDir::CFileNameModuleDir(const string& strFileName): CFileName(strFileName) {
    SetBaseDir(GetModuleDir());
}

CFileNameSystemDir::CFileNameSystemDir(): CFileName(GetSystemDir()) {
}

CFileNameSystemDir::CFileNameSystemDir(const string& strFileName): CFileName(strFileName) {
    SetBaseDir(GetSystemDir());
}

CFileNameNoDir::CFileNameNoDir(const string& strFileName): CFileName(strFileName) {
    SetBaseDir(string());
}

CFileName::CFileName(const string& strFileName, const string& strBaseDir): m_strName(strFileName) {
    Themis_DebugAssert(!CFileName(strBaseDir).IsRelativePath());
    Themis_DebugAssert(CFileName(strBaseDir).IsDirPath());
    SetBaseDir(strBaseDir);
}

CFileName::CFileName(const string& strFileName, const CFileName& oBaseDir): m_strName(strFileName) {
    Themis_DebugAssert(!oBaseDir.IsRelativePath());
    Themis_DebugAssert(oBaseDir.IsDirPath());
    SetBaseDir(oBaseDir.Get());
}

CFileName::~CFileName() {
}

string CFileName::GetSuffix() const {
    int nDot = GetLastDotPos(m_strName);
    return (nDot < 0)? string(): m_strName.Substring(nDot + 1);
}

string CFileName::GetSuffixLc() const {
    int nDot = GetLastDotPos(m_strName);
    return (nDot < 0)? string(): m_strName.Substring(nDot + 1).ToLower();
}

string CFileName::GetPrefix() const {
    string strName = GetPureName();
    int nDot = GetLastDotPos(strName);
    return (nDot < 0)? strName: strName.Substring(0, nDot);
}

string CFileName::GetPureName() const {
    int nSep = GetLastSeparatorPos(m_strName);
    return (nSep < 0) ? m_strName : m_strName.Substring(nSep + 1);
}

string CFileName::GetDir() const {
    int nSep = GetLastSeparatorPos(m_strName);
    return (nSep < 0) ? string(): m_strName.Substring(0, nSep + 1);
}

string CFileName::GetDirName() const {
    string str = GetDir();
    if (string::IsNullOrEmpty(str) || (str.GetLength() == 3 && str[1] == L':')) {
        return string();
    } else {
        return CFileName(RemoveDirSeparator(str)).GetPureName();
        //蔲ۂC邯ǂƂ肠B
    }
}

string CFileName::GetParentDir() const {
    if (string::IsNullOrEmpty(m_strName) || IsRootPath()) {
        return string();
    }
    return CFileName(RemoveDirSeparator(m_strName)).GetDir();
}

wchar_t CFileName::GetDriveLetter() const {
    if (IsNetworkPath()) {
        return L'/';
    } else if (IsFullPath()) {
        return m_strName[0];
    } else {
        throw RuntimeException(m_strName + L" ̃hCu^[̎擾Ɏs܂");
    }
}


string CFileName::GetRelativeName(const CFileName& oDirName, int nMaxUp) const {
    Themis_DebugAssert(!this->IsRelativePath());
    Themis_DebugAssert(!oDirName.IsRelativePath());
    Themis_DebugAssert(oDirName.IsDirPath());
    string strDst = oDirName.Get();
    string strSrc = Get();

    //JgfBNgƓfBNgȕJEg
    int nMatchDirPos = 0;
    int n = Min(strDst.GetLength(), strSrc.GetLength());
    for (int i = 0; i < n; ) {
        if (strDst[i] == strSrc[i]) {
            if (IsYen(strDst[i])) {
                nMatchDirPos = i;
            }
            i++;
        } else {
            break;
        }
    }
    if (nMatchDirPos <= 0) {    // fBNgȕSȂ
        return Get();           // tpX̂̂B
    }

    // L".."̐(vȍ~\\̐)JEg
    int nUpCount = 0;
    for (int j = nMatchDirPos + 1; j < strDst.GetLength(); j++) {
        if (IsYen(strDst[j])) {
            nUpCount++;
        }
    }
    if (nMaxUp < 0 || nUpCount < nMaxUp) {
        string strRel;
        for (int k = 0; k < nUpCount; k++) {
            strRel += L"..\\";
        }
        return (strRel + strSrc.Substring(nMatchDirPos + 1));
    } else {
        // ..\\̐ȂtpXŁB
        return Get();
    }
}

string CFileName::AddDirSeparator() const {
    return AddDirSeparator(m_strName);
}

string CFileName::RemoveDirSeparator() const {
    return RemoveDirSeparator(m_strName);
}

string CFileName::RemoveDirSeparatorForce() const {
    return RemoveDirSeparatorForce(m_strName);
}

CFileName CFileName::SetSuffix(const string& strSuffix) const {
    int nDot = GetLastDotPos(m_strName);
    if (0 < nDot) {                 //_
        if (string::IsNullOrEmpty(strSuffix)) {    //gq
            return CFileName(m_strName.Substring(0, nDot));
        } else {                      //gqA
            return CFileName(m_strName.Substring(0, nDot + 1) + strSuffix);
        }
    } else {                          //_Ȃ
        if (string::IsNullOrEmpty(strSuffix)) {    //gq
            return *this;
        } else {                      //gqA
            return CFileName(m_strName + L"." + strSuffix);
        }
    }
}

CFileName CFileName::SetPrefix(const string& strPrefix) const {
    return CFileName(GetDir() + strPrefix + L"." + GetSuffix());
}

CFileName CFileName::SetPureName(const string& strPureName) const {
    return CFileName(GetDir() + strPureName);
}

CFileName CFileName::SetDir(const CFileName& oDir) const {
    Themis_DebugAssert(!oDir.IsRelativePath());
    Themis_DebugAssert(oDir.IsDirPath());
    return CFileName(oDir.Get() + GetPureName());
}


bool CFileName::operator==(const CFileName& r) const {
    Themis_DebugAssert(IsEmpty() || !IsRelativePath());
    Themis_DebugAssert(!r.IsRelativePath());
    return Get().ToLower() == r.Get().ToLower();
}

bool CFileName::operator<(const CFileName& r) const {
    Themis_DebugAssert(IsEmpty() || !IsRelativePath());
    Themis_DebugAssert(!r.IsRelativePath());
    return Get().ToLower() < r.Get().ToLower();
}


void CFileName::SetBaseDir(const string& strBaseDir) {
    string strName = Environment::ExpandEnvironmentVariables(m_strName);
    //fBNgAtpXlbg[NpXɂȂĂI
    if (string::IsNullOrEmpty(strBaseDir) || !CFileName(strName).IsRelativePath()) {
        m_strName = SlashToYen(strName);
    } else {
        //΃pX̓WJ
        Themis_DebugAssert(CFileName(strName).IsRelativePath());
        StringBuilder builder = AddDirSeparator(strBaseDir);
        for (int i = 0, n = strName.GetLength(); i < n; ) {
            if (IsSeparator(strName[i])) {
                i += 1;
                if (i != 0) {
                    builder.Append(L'\\');
                }
            } else if (strName[i] == L'.') {
                if (3 <= strName.GetLength() - i && strName[i + 1] == L'.' && IsSeparator(strName[i + 2])) {
                    //̃fBNg
                    int nYenPos = GetLastSeparatorPos(RemoveDirSeparator(builder.ToString()));
                    if (nYenPos < 0) {
                        Themis_DebugWarning(L"sȃt@C");
                        m_strName = strName;
                        break;
                    }
                    builder.Remove(nYenPos + 1, 1);
                    i += 3;
                } else if (2 <= strName.GetLength() - i && IsSeparator(strName[i + 1])) {
                    //Ƃ肠邾
                    i += 2;
                } else {
                    // . Ȃ̂ŃRs[
                    builder.Append(L'.');
                    i += 1;
                }
            } else {
                builder.Append(strName[i]);
                i += 1;
            }
        }
        m_strName = builder.ToString();
    }
}

string CFileName::AddDirSeparator(const string& str) {
    return (string::IsNullOrEmpty(str) || IsSeparator(str[str.GetLength() - 1])) ? str : str + L"\\";
}

string CFileName::RemoveDirSeparator(const string& str) {
    if (string::IsNullOrEmpty(str) || !IsSeparator(str[str.GetLength() - 1]) || CFileName(str).IsRootPath()) {
        return str; // 󂩃tH_؂薳 "C:\\" ƂȂ炻̂܂ܕԂ
    }
    return str.Substring(0, str.GetLength() - 1);
}

string CFileName::RemoveDirSeparatorForce(const string& str) {
    if (string::IsNullOrEmpty(str) || !IsSeparator(str[str.GetLength() - 1])) {
        return str; // 󂩃tH_؂薳Ȃ炻̂܂ܕԂ
    }
    return str.Substring(0, str.GetLength() - 1);
}

string CFileName::SlashToYen(const string& str) {
    return str.Replace(L'/', L'\\');
}

int CFileName::GetLastSeparatorPos(const string& str) {
    int nPos = -1;
    for (int i = 0, n = str.GetLength(); i < n; ) {
        if (IsSeparator(str[i])) {
            nPos = static_cast<int>(i++);
        } else {
            i++;
        }
    }
    return nPos;
}

int CFileName::GetLastDotPos(const string& str) {
    int nPos = -1;
    for (int i = 0, n = str.GetLength(); i < n; ) {
        if (str[i] == L'.') {
            nPos = static_cast<int>(i++);
        } else if (IsSeparator(str[i])) {
            nPos = -1;      //fBNg؂肠炻ȑÔ͖
            i++;
        } else {
            i++;
        }
    }
    return nPos;
}

}
