#pragma once
#include "Windows/NativeString.h"
#include "String/Encoding.h"
namespace Themis {

/**
    t@CB

    Tv
@code

    CFindFile oFindFile(strDir + L"*");
    do {
        if (!oFindFile.IsDot()) {
            //FXB
        }
    } while (oFindFile.ToNext());

@endcode
*/
class CFindFile : private boost::noncopyable {
public:
    /// ŏ̃t@Cݒ(ChJ[hgp\)
    CFindFile(const string& strFirstName)
        : m_hFile(INVALID_HANDLE_VALUE), useWideChar(NativeString::IsUnicode()) {
        m_hFile = useWideChar ?
            ::FindFirstFileW(strFirstName.ToCharPtr(), &findData.W) :
            ::FindFirstFileA(Encoding::GetDefault().GetBytes(strFirstName).c_str(), &findData.A);
        if (m_hFile == INVALID_HANDLE_VALUE) {
            throw Win32Exception(strFirstName + L" ̌Ɏs");
        }
        m_strFirstNameDir = CFileName(strFirstName).GetDir();
        m_strFirstNamePureName = CFileName(strFirstName).GetPureName();
    }
    ~CFindFile() {
        if (m_hFile != INVALID_HANDLE_VALUE) {
            ::FindClose(m_hFile);
        }
    }

    /// ̃t@C
    bool ToNext() {
        BOOL result = useWideChar ?
            ::FindNextFileW(m_hFile, &findData.W) :
            ::FindNextFileA(m_hFile, &findData.A);
        return result != FALSE;
    }

    /// ݂̃t@C̑擾
    //
    // FILE_ATTRIBUTE_ARCHIVE
    // FILE_ATTRIBUTE_COMPRESSED
    // FILE_ATTRIBUTE_DIRECTORY
    // FILE_ATTRIBUTE_ENCRYPTED
    // FILE_ATTRIBUTE_HIDDEN
    // FILE_ATTRIBUTE_NORMAL
    // FILE_ATTRIBUTE_OFFLINE
    // FILE_ATTRIBUTE_READONLY
    // FILE_ATTRIBUTE_REPARSE_POINT
    // FILE_ATTRIBUTE_SPARSE_FILE
    // FILE_ATTRIBUTE_SYSTEM
    // FILE_ATTRIBUTE_TEMPORARY
    //
    DWORD GetAttributes() const {
        return useWideChar ?
            findData.W.dwFileAttributes :
            findData.A.dwFileAttributes;
    }
    /// ݂̂fBNgȂ̂ǂ
    bool IsDir() const { return (GetAttributes() & FILE_ATTRIBUTE_DIRECTORY) != 0; }
    /// u.v, u..vȏꍇtrueԂ
    bool IsDot() const {
        if (!IsDir()) return false;
        if (useWideChar) {
            if (findData.W.cFileName[0] == L'.') {
                Themis_DebugAssert(findData.W.cFileName[1] != L'\0' || findData.W.cFileName[1] != L'.');
                return true;
            }
        } else {
            if (findData.A.cFileName[0] == '.') {
                Themis_DebugAssert(findData.A.cFileName[1] != '\0' || findData.A.cFileName[1] != '.');
                return true;
            }
        }
        return false;
    }
    /// ݂̃t@C̃TCY擾
    size_t GetSize() const {
        return useWideChar ? 
            findData.W.nFileSizeLow :
            findData.A.nFileSizeLow;
    }
    /// ݂̃t@C̃TCY擾̂Q
    uint64_t GetSize64() const {
        return useWideChar ? 
            ((uint64_t(findData.W.nFileSizeHigh) << 32) | findData.W.nFileSizeLow) :
            ((uint64_t(findData.A.nFileSizeHigh) << 32) | findData.A.nFileSizeLow);
    }
    /// ݂̃t@C(tpXt)擾BɃfBNgLƂ͖
    string GetFullName() const {
        return m_strFirstNameDir + GetName();
    }
    /// ݂̃t@C(tpXt)擾BKvɉĖɃfBNgLtB
    CFileName GetFullFileName() const {
        return CFileName(IsDir() ? GetFullName() + L"\\" : GetFullName());
    }
    /// t@CEfBNg̎擾
    string GetName() const {
        return useWideChar ? 
            string(findData.W.cFileName) :
            Encoding::GetDefault().GetString(findData.A.cFileName);
    }
    /// t@C̍XV̎擾
    const FILETIME& GetLastUpdate() const  {
        return useWideChar ?
            findData.W.ftLastWriteTime :
            findData.A.ftLastWriteTime;
    }
    /// FIND_DATA̒ڎ擾
 // const WIN32_FIND_DATA& GetFindData() const { return findData.; }

    /// ŏ̃t@C̎擾
    string GetFiastName() const { return m_strFirstNameDir + m_strFirstNamePureName; }

private:
    string              m_strFirstNameDir;
    string              m_strFirstNamePureName;
    HANDLE              m_hFile;
    union FindData {
        WIN32_FIND_DATAA A;
        WIN32_FIND_DATAW W;
    };
    FindData findData;
    const bool useWideChar;
};

}
