#include "Font.h"
#include "../Bitmap/PlaneUtility.h"
#include "../../Char.h"
#include "../../Thread/ThreadSpecificObject.h"
#include "../../String/Encoding.h"
#include "../../Utility/Algorithm.h"
namespace Themis {

CFont::CFont(const CFontStyle& rFontStyle): m_fs(rFontStyle), m_hFont(NULL) {
    LOGFONT lf = rFontStyle.GetLogFont();
 // lf.lfQuality = ANTIALIASED_QUALITY;
    m_hFont = ::CreateFontIndirect(&lf);
}

CFont::~CFont() {
    if (m_hFont != NULL) {
        ::DeleteObject(m_hFont);
    }
}

/// ̏c擾
int CFont::GetLineHeight(int nLine) const {
    Themis_DebugAssert(0 < nLine);
    const int nShadowGap = GetShadowGapSize();
    return (m_fs.GetHeight() + nShadowGap) * nLine - nShadowGap + m_fs.GetPadding().top + m_fs.GetPadding().bottom;
}

/// ̑傫擾
CSize CFont::GetSize(const string& text) {
    List<string> vct = text.TrimEnd().Split(Array<string>(L"\r\n", L"\n", L"\r"));
    HDC hDst = ::CreateCompatibleDC(NULL);
    HFONT hOldFnt = reinterpret_cast<HFONT>(::SelectObject(hDst, m_hFont));
    CSize oSize(0, 0);
    if (vct.size() == 1) { // Psȏꍇ
        std::string strText = Encoding::GetDefault().GetBytes(text);
        BOOL bRet = ::GetTextExtentPoint32(hDst, strText.c_str(), static_cast<int>(strText.length()), &oSize);
        if (bRet == FALSE) {
            Win32Exception o(L"`敶̐@̎擾Ɏs");
            ::SelectObject(hDst, hOldFnt);
            ::DeleteDC(hDst);
            throw o;
        }
        const int nGap = GetShadowGapSize();
        oSize.cx += nGap;
        oSize.cy += nGap;
    } else if (!vct.empty()) { // sȏꍇ
        const int nShadowGap = GetShadowGapSize();
        const int nLineHeight = m_fs.GetHeight() + nShadowGap + m_fs.GetLineGap();
        for (size_t i = 0, n = vct.size(); i < n; i++) {
            CSize szTmp;
            std::string mb = Encoding::GetDefault().GetBytes(vct[i]);
            BOOL bRet = ::GetTextExtentPoint32(hDst, mb.c_str(), static_cast<int>(mb.length()), &szTmp);
            if (bRet == FALSE) {
                Win32Exception o(L"`敶̐@̎擾Ɏs");
                ::SelectObject(hDst, hOldFnt);
                ::DeleteDC(hDst);
                throw o;
            }
            if (oSize.cx < szTmp.cx) {
                oSize.cx = szTmp.cx;
            }
            oSize.cy += nLineHeight;
        }
        oSize.cx += nShadowGap;
        oSize.cy -= m_fs.GetLineGap(); // Ō̂Ƃ͗vȂ
    }
    ::SelectObject(hDst, hOldFnt);
    ::DeleteDC(hDst);
    return oSize + CSize(m_fs.GetPadding().left + m_fs.GetPadding().right, m_fs.GetPadding().top + m_fs.GetPadding().bottom);
}

namespace {
    CDib32& GetTempDib() {
        static CThreadSpecificObject<CDib32> oTmpDib;
        return oTmpDib.Get();
    }
}
/// ̕`
void CFont::DrawText(CPlaneEx& rDst, const CPoint& pt, const string& text) {
    if (m_fs.IsAntiAlias()) {
        if (rDst.GetBpp() == 32) {
            InnerDrawTextAA(static_cast<CPlane32&>(rDst), pt, text);
        } else {
            Themis_DebugLog(L"32BPPȃv[ւ̂񂿂蕶͔ۂł");
            CDib32& oDib = GetTempDib();
            CSize sz = GetSize(text) + CSize(16, 16);
            oDib.SetSize(sz);
            oDib.Blt(rDst, CPoint(), CRect(pt - CSize(8, 8), sz));
            InnerDrawTextAA(oDib, CPoint(8, 8), text);
            rDst.Blt(oDib, pt - CSize(8, 8), CRect(sz));
        }
    } else {
        if (rDst.IsGdiSupport()) {
            InnerDrawText(CPlane::CGdiLock(rDst), pt, text);
        } else {
            Themis_DebugLog(L"HDCȂv[ւ̔񂠂񂿂蕶͔ۂł");
            CDib32& oDib = GetTempDib();
            CSize sz = GetSize(text) + CSize(16, 16);
            oDib.SetSize(sz);
            oDib.Blt(rDst, CPoint(), CRect(pt - CSize(8, 8), sz));
            InnerDrawText(CPlane::CGdiLock(oDib), CPoint(8, 8), text);
            rDst.Blt(oDib, pt - CSize(8, 8), CRect(sz));
        }
    }
}

/// ̕`悻̂Q
void CFont::DrawText(HDC hDst, const CPoint& pt, const string& text) {
    if (m_fs.IsAntiAlias()) {
        Themis_DebugLog(L"HDCւ̂񂿂蕶͔ۂł");
        CDib32& oDib = GetTempDib();
        CSize sz = GetSize(text) + CSize(16, 16);
        oDib.SetSize(sz);
        CDC(CPlane::CGdiLock(oDib)).Blt(hDst, CPoint(), CRect(pt - CSize(8, 8), sz));
        InnerDrawTextAA(oDib, CPoint(8, 8), text);
        CDC(hDst).Blt(CPlane::CGdiLock(oDib), pt - CSize(8, 8), CRect(sz));
    } else {
        InnerDrawText(hDst, pt, text);
    }
}

void CFont::InnerDrawText(HDC hDst, const CPoint& pt, const string& text) {
    // wi`
    if (m_fs.IsBg()) {
        CDC(hDst).Fill(m_fs.GetBgColor(), CRect(pt, GetSize(text)));
    }
    // es̕`
    HFONT hFontOld = (HFONT)::SelectObject(hDst, m_hFont);
    int iBkMode = ::SetBkMode(hDst, TRANSPARENT);
    List<string> vct = text.TrimEnd().Split(Array<string>(L"\r\n", L"\n", L"\r"));
    CPoint pt2 = pt;
    pt2.x += m_fs.GetPadding().left;
    pt2.y += m_fs.GetPadding().top;
    const int nLineHeight = m_fs.GetHeight() + GetShadowGapSize() + m_fs.GetLineGap();
    for (size_t i = 0, n = vct.size(); i < n; i++) {
        InnerDrawTextLine(hDst, pt2, vct[i]);
        pt2.y += nLineHeight;
    }
    ::SetBkMode(hDst, iBkMode);
    ::SelectObject(hDst, hFontOld);
}

void CFont::InnerDrawTextAA(CPlane32& rDst, const CPoint& pt, const string& text) {
    // wi`
    if (m_fs.IsBg()) {
        rDst.Fill(m_fs.GetBgColor(), CRect(pt, GetSize(text)));
    }
    // es̕`
    vector<string> vct = text.TrimEnd().Split(Array<string>(L"\r\n", L"\n", L"\r"));
    while (!vct.empty() && string::IsNullOrEmpty(vct[vct.size() - 1])) {
        vct.erase(vct.begin() + (vct.size() - 1));
    }
    CPoint pt2 = pt;
    pt2.x += m_fs.GetPadding().left;
    pt2.y += m_fs.GetPadding().top;
    const int nLineHeight = m_fs.GetHeight() + GetShadowGapSize() + m_fs.GetLineGap();
    for (size_t i = 0, n = vct.size(); i < n; i++) {
        InnerDrawTextLineAA(rDst, pt2, vct[i]);
        pt2.y += nLineHeight;
    }
}

void CFont::InnerDrawTextLine(HDC hDst, const CPoint& pt, const string& text) {
    std::string strText = Encoding::GetDefault().GetBytes(text);
    // e̕`Ƃ
    COLORREF rgbOldColor;
    if (m_fs.GetShadowType() == 0) {
        //t`eȂ
        rgbOldColor = ::SetTextColor(hDst, m_fs.GetColor());
        ::TextOut(hDst, pt.x, pt.y, strText.c_str(), static_cast<int>(strText.length()));
    } else {
        //t`ore
        int nGap = m_fs.GetShadowGap();
        int nGap2 = nGap * 2;
        if (m_fs.GetShadowType() == 1) {
            //e(E)
            rgbOldColor = ::SetTextColor(hDst, m_fs.GetShadowColor());
            ::TextOut(hDst, pt.x + nGap, pt.y + nGap, strText.c_str(), static_cast<int>(strText.length()));
            ::SetTextColor(hDst, m_fs.GetColor());
            ::TextOut(hDst, pt.x, pt.y, strText.c_str(), static_cast<int>(strText.length()));
        } else {
            //t`(EEEEEE㉺E)
            rgbOldColor = ::SetTextColor(hDst, m_fs.GetShadowColor());
            ::TextOut(hDst, pt.x, pt.y, strText.c_str(), static_cast<int>(strText.length()));    //
            ::TextOut(hDst, pt.x + nGap2, pt.y, strText.c_str(), static_cast<int>(strText.length()));    //E
            ::TextOut(hDst, pt.x, pt.y + nGap2, strText.c_str(), static_cast<int>(strText.length()));    //
            ::TextOut(hDst, pt.x + nGap2, pt.y + nGap2, strText.c_str(), static_cast<int>(strText.length()));    //E
            ::TextOut(hDst, pt.x + nGap, pt.y, strText.c_str(), static_cast<int>(strText.length()));    //
            ::TextOut(hDst, pt.x + nGap, pt.y + nGap2, strText.c_str(), static_cast<int>(strText.length()));    //
            ::TextOut(hDst, pt.x, pt.y + nGap, strText.c_str(), static_cast<int>(strText.length()));    //
            ::TextOut(hDst, pt.x + nGap2, pt.y + nGap, strText.c_str(), static_cast<int>(strText.length()));    //E
            ::SetTextColor(hDst, m_fs.GetColor());
            ::TextOut(hDst, pt.x + nGap, pt.y + nGap, strText.c_str(), static_cast<int>(strText.length()));
        }
    }
    ::SetTextColor(hDst, rgbOldColor);
}

void CFont::InnerDrawTextLineAA(CPlane32& rDst, const CPoint& pt, const string& strText) {
    // t`Ee̕`
    if (m_fs.GetShadowType() == 0) {
        //t`eȂ
        InnerTextOutLineAA(rDst, pt, strText, m_fs.GetColor());
    } else {
        //t`ore
        int nGap = m_fs.GetShadowGap();
        int nGap2 = nGap * 2;
        if (m_fs.GetShadowType() == 1) {
            //e(E)
            InnerTextOutLineAA(rDst, CPoint(pt.x + nGap, pt.y + nGap), strText, m_fs.GetShadowColor());
            InnerTextOutLineAA(rDst, pt, strText, m_fs.GetColor());
        } else {
            //t`(EEEEEE㉺E)
            InnerTextOutLineAA(rDst, pt, strText, m_fs.GetShadowColor());    //
            InnerTextOutLineAA(rDst, CPoint(pt.x + nGap2, pt.y), strText, m_fs.GetShadowColor());    //E
            InnerTextOutLineAA(rDst, CPoint(pt.x, pt.y + nGap2), strText, m_fs.GetShadowColor());    //
            InnerTextOutLineAA(rDst, CPoint(pt.x + nGap2, pt.y + nGap2), strText, m_fs.GetShadowColor());    //E
            InnerTextOutLineAA(rDst, CPoint(pt.x + nGap, pt.y), strText, m_fs.GetShadowColor());    //
            InnerTextOutLineAA(rDst, CPoint(pt.x + nGap, pt.y + nGap2), strText, m_fs.GetShadowColor());    //
            InnerTextOutLineAA(rDst, CPoint(pt.x, pt.y + nGap), strText, m_fs.GetShadowColor());    //
            InnerTextOutLineAA(rDst, CPoint(pt.x + nGap2, pt.y + nGap), strText, m_fs.GetShadowColor());    //E
            InnerTextOutLineAA(rDst, CPoint(pt.x + nGap, pt.y + nGap), strText, m_fs.GetColor());
        }
    }
}

namespace {
    FIXED MakeFixed(uint32_t dw) { return *(FIXED*)&dw; }
}
/// ̕`
void CFont::InnerTextOutLineAA(CPlane32& rDst, const CPoint& pt, const string& text, COLORREF rgbFont) {
    HDC hdc = ::CreateCompatibleDC(NULL);
    HFONT hFontOld = (HFONT)::SelectObject(hdc, m_hFont);
    Themis_PlaneLock4(rDst, pt, pDst, nDstPitch);
    Themis_PlaneGetRGB(rgbFont, nTextR, nTextG, nTextB);
    uint32_t dwTextBGR = Detail::BGR32(nTextR, nTextG, nTextB);
    static CThreadSpecificObject<vector<uint8_t> > oGlyphBuf;
    vector<uint8_t>& rGlyphBuf = oGlyphBuf.Get();
    if (rGlyphBuf.empty()) { // Ô߂PoCg͊mۂƂ
        rGlyphBuf.resize(1);
    }
    // tHg擾
    TEXTMETRIC tm;
    ::GetTextMetrics(hdc, &tm);
    // ]sH
    //  1  0
    //  0  1
    MAT2 mt;
    mt.eM11 = MakeFixed(0x10000);
    mt.eM12 = MakeFixed(0);
    mt.eM21 = mt.eM12;
    mt.eM22 = mt.eM11;
    std::string strText = Encoding::GetDefault().GetBytes(text);
    for (size_t i = 0, n = strText.length(); i < n; ) {
        // Po
        UINT uChar;
        if (CChar::Is2Byte(strText[i])) {
            uChar = static_cast<uint8_t>(strText[i]);
            uChar <<= 8;
            uChar |= static_cast<uint8_t>(strText[i + 1]);
            i += 2;
        } else {
            uChar = static_cast<uint8_t>(strText[i++]);
        }
        // obt@TCY̎擾
        GLYPHMETRICS gm;
        DWORD dwSize = ::GetGlyphOutline(hdc, uChar, GGO_GRAY8_BITMAP, &gm, 0, NULL, &mt);
        const int nGlyphPitch = (gm.gmBlackBoxX + 3) & ~3;
     // Themis_DebugAssert(nGlyphPitch * gm.gmBlackBoxY <= dwSize);
        // obt@
        if (rGlyphBuf.size() < dwSize) {
            rGlyphBuf.resize(dwSize);
        }
        Themis_DebugAssert(nGlyphPitch * gm.gmBlackBoxY <= rGlyphBuf.size());
        // A`GCAXtHgf[^擾
        DWORD dwRet = ::GetGlyphOutline(hdc, uChar, GGO_GRAY8_BITMAP, &gm, dwSize, &rGlyphBuf[0], &mt);
        Themis_DebugAssertApi(dwRet != GDI_ERROR);
        // `[
        uint32_t* pDst2 = pDst + gm.gmptGlyphOrigin.x + nDstPitch * (tm.tmAscent - gm.gmptGlyphOrigin.y);
        const int nBoxCY = Min<int>(gm.gmBlackBoxY, rDst.GetSize().cy - pt.y);
        const int nBoxCX = Min<int>(gm.gmBlackBoxX, rDst.GetSize().cx - pt.x);
        for (int y = 0; y < nBoxCY; y++) {
            uint8_t* pbyGlyph = &rGlyphBuf[nGlyphPitch * y];
            for (int x = 0; x < nBoxCX; x++) {
                Themis_DebugAssert(0 <= pbyGlyph[x] && pbyGlyph[x] <= 64); // Ôassert
                const int t = pbyGlyph[x];
                if (t != 0) {
                    if (t == 64) { // 100`F
                        pDst2[x] = dwTextBGR;
                    } else { // `O̐Fƃuh
                        Themis_PlaneGetBGR32(pDst2[x], dr, dg, db);
                        const int r = (int(nTextR - dr) * t / 64) + dr;
                        const int g = (int(nTextG - dg) * t / 64) + dg;
                        const int b = (int(nTextB - db) * t / 64) + db;
                        Themis_PlaneCheckByteRGB(r, g, b); // ͈͓ǂassert
                        pDst2[x] = Detail::BGR32(r, g, b);
                    }
                }
            }
            pDst2 += nDstPitch;
        }
        // ̕̕`ʒuֈړ
        pDst += gm.gmCellIncX; //gm.gmBlackBoxX;
    }
    ::SelectObject(hdc, hFontOld);
    ::DeleteDC(hdc);
}

/// e≏̑傫擾
int CFont::GetShadowGapSize() const {
    switch (GetStyle().GetShadowType()) {
    case 0: return 0;
    case 1: return GetStyle().GetShadowGap();
    case 2: return GetStyle().GetShadowGap() * 2;
    default: Themis_DebugTrespass();
    }
    return 0;
}

}
