#include "stdafx.h"
#include "Png.h"
#include "../File.h"
#include "../Graphics.h"
#include "../String/Encoding.h"
#if Themis_ConfigPngLib
# if !Themis_ConfigZLib
#  error libpng needs 'zlib'.
# endif
# include "../../../external/libpng.h"
# ifdef NDEBUG
#  pragma comment(lib, "libpng_mr.lib")
# else
#  pragma comment(lib, "libpng_md.lib")
# endif
# ifdef NDEBUG
#  pragma comment(lib, "zlib_mr.lib")
# else
#  pragma comment(lib, "zlib_md.lib")
# endif
#endif
namespace Themis {

#if Themis_ConfigPngLib

namespace {
    void PNGAPI my_png_error PNGARG((png_structp, png_const_charp)) {
        throw ImageLoadException();
    }
    void PNGAPI my_png_warn PNGARG((png_structp, png_const_charp msg)) {
        Themis_DebugLog(msg);
    }
}

CPngLoader::CPngLoader(): m_png_ptr(NULL), m_info_ptr(NULL), m_bAlpha(false) {
    m_png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, &my_png_error, &my_png_warn);
    m_info_ptr = png_create_info_struct(m_png_ptr);
}

CPngLoader::~CPngLoader() {
    if (m_png_ptr != NULL && m_info_ptr != NULL) {
        png_destroy_read_struct(&m_png_ptr, &m_info_ptr, NULL);
    } else if (m_png_ptr != NULL) {
        png_destroy_read_struct(&m_png_ptr, NULL, NULL);
    }
}

namespace {
    void PngReadFunction(png_struct* png, png_bytep pBuf, png_size_t nSize) {
        CPngLoader::SFileData* pFile = static_cast<CPngLoader::SFileData*>(png_get_io_ptr(png));
        const vector<uint8_t>& fileMemory = pFile->fileData->GetMemory();
        size_t nNewPos = pFile->nPos + nSize;
        if (nNewPos <= fileMemory.size()) {
            memcpy(pBuf, &fileMemory[0] + pFile->nPos, nSize);
            pFile->nPos = nNewPos;
        } else { // t@CTCYĂꍇBG[ԂȂ̂łƂ肠OŖ߂Ă݂BB
            size_t nReadSize = fileMemory.size() - pFile->nPos;
            memcpy(pBuf, &fileMemory[0] + pFile->nPos, nReadSize);
            memset(pBuf + nReadSize, 0x00, nSize - nReadSize);
        }
    }
}
bool CPngLoader::Load(const std::tr1::shared_ptr<CFileData>& fileData, bool bRenderOnlyAtOnce) {
    try {
        if (m_png_ptr == NULL || m_info_ptr == NULL) {
            return false;
        }
        png_byte buf[8];
        memcpy(buf, &fileData->GetMemory()[0], 8);
        if (png_check_sig(buf, 8) == 0) {
            return false;
        }
     /*
        png_init_io(m_png_ptr, fp);
     /*/
        m_pFileData.reset(new SFileData(fileData, 8));
        png_set_read_fn(m_png_ptr, m_pFileData.get(), &PngReadFunction);
     //*/
        png_set_sig_bytes(m_png_ptr, 8);
        png_read_info(m_png_ptr, m_info_ptr);
        png_uint_32 nImgW, nImgH;
        int nBitDepth;
        int nColorType;
        png_get_IHDR(m_png_ptr, m_info_ptr, &nImgW, &nImgH, &nBitDepth, &nColorType, NULL, NULL, NULL);
        // `̕ϊ
        if (nColorType == PNG_COLOR_TYPE_PALETTE ||
            (nColorType == PNG_COLOR_TYPE_GRAY && nBitDepth < 8) ||
            (png_get_valid(m_png_ptr, m_info_ptr, PNG_INFO_tRNS))) {
            png_set_expand(m_png_ptr);
        }
        if (nBitDepth == 16) {
            png_set_strip_16(m_png_ptr);
        }
        if (nColorType == PNG_COLOR_TYPE_GRAY ||
            nColorType == PNG_COLOR_TYPE_GRAY_ALPHA) {
            png_set_gray_to_rgb(m_png_ptr);
        }
        // }XN
        if (nColorType & PNG_COLOR_MASK_ALPHA) {
            m_bAlpha = true;
        } else {
            // }XNȂ0xffŖ߂
            png_set_filler(m_png_ptr, 0xff, PNG_FILLER_AFTER);
        }
        // K}l̐ݒ
        double dGamma = 0.0;
        if (png_get_gAMA(m_png_ptr, m_info_ptr, &dGamma)) {
            png_set_gamma(m_png_ptr, 2.2, dGamma);
        }
        // ׂĂ̐ݒo^{߂ďoB
        png_read_update_info(m_png_ptr, m_info_ptr);
        png_get_IHDR(m_png_ptr, m_info_ptr, &nImgW, &nImgH, &nBitDepth, &nColorType, NULL, NULL, NULL);
        m_szImage = CSize(nImgW, nImgH);
        // `lĂȂBPP == 32Ȃ̂mFƂB
        Themis_DebugAssert((m_bAlpha && png_get_channels(m_png_ptr, m_info_ptr) == 4) || !m_bAlpha);
        if (m_bAlpha && png_get_channels(m_png_ptr, m_info_ptr) != 4) {
            m_bAlpha = false; // Ô߁BB
        }
        // ꎞv[֕`
        if (!bRenderOnlyAtOnce) {
            const int nImgRowBytes = png_get_rowbytes(m_png_ptr, m_info_ptr);
            const int nImgChannels = png_get_channels(m_png_ptr, m_info_ptr);
            m_pPlane.reset(new CMemoryPlane(nImgChannels * 8));
            m_pPlane->SetSize(m_szImage);
            Themis_PlaneLock1(*m_pPlane, CPoint(), p, nPitch);
        /*
            boost::scoped_array<png_bytep> apPngRow(new png_bytep[nImgH]);
            for (png_uint_32 y = 0; y < nImgH; y++) {
                apPngRow[y] = static_cast<png_bytep>(p);
                p += nPitch;
            }
            png_set_bgr(m_png_ptr);
            png_read_image(m_png_ptr, apPngRow.get());
            png_read_end(m_png_ptr, m_info_ptr);
        /*/
            png_set_bgr(m_png_ptr);
            const int nPass = png_set_interlace_handling(m_png_ptr);
            const int nImgHeight = m_png_ptr->num_rows = m_png_ptr->height; // Ȃ񂩂悭񂯂png_read_image̐^
            for (int i = 0; i < nPass; i++) {
                uint8_t* p2 = p;
                for (int y = 0; y < nImgHeight; y++) {
                    png_read_row(m_png_ptr, p2, NULL);
                    p2 += nPitch;
                }
            }
            png_read_end(m_png_ptr, m_info_ptr);
        //*/
        }
        return true;
    } catch (std::exception&) {
        m_pFileData.reset();
        return false;
    }
}

CSize CPngLoader::GetSize() const {
    return m_szImage;
}

void CPngLoader::Render(CPlane& dst, const CPoint& pt, bool bRenderOnlyAtOnce) {
    try {
        if (bRenderOnlyAtOnce) {
            const int nImgRowBytes = png_get_rowbytes(m_png_ptr, m_info_ptr);
            const int nImgChannels = png_get_channels(m_png_ptr, m_info_ptr);
            if (dst.GetBpp() == nImgChannels * 8) {
                Themis_PlaneLock1(dst, pt, p, nPitch);
                Themis_PlaneGetSize(m_szImage, nSrcW, nSrcH);
                png_set_bgr(m_png_ptr);
                const int nPass = png_set_interlace_handling(m_png_ptr);
                const int nImgHeight = m_png_ptr->num_rows = m_png_ptr->height; // Ȃ񂩂悭񂯂png_read_image̐^
                for (int i = 0; i < nPass; i++) {
                    uint8_t* p2 = p;
                    for (int y = 0; y < nImgHeight; y++) {
                        png_read_row(m_png_ptr, p2, NULL);
                        p2 += nPitch;
                    }
                }
                png_read_end(m_png_ptr, m_info_ptr);
            } else {
                Themis_DebugWarning(L"ۂPNGǂݍ݂ł");
            }
        } else {
            dst.Blt(*m_pPlane, pt, CRect(dst.GetSize()));
        }
    } catch (std::exception&) {
        throw ImageLoadException();
    }
}

string CPngLoader::GetCopyright() {
    return Encoding::GetDefault().GetString(png_get_copyright(NULL) + 1); //ŏ\n̂ŁBB
}


CPngSaver::CPngSaver(): m_png_ptr(NULL), m_info_ptr(NULL) {
    m_png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, &my_png_error, &my_png_warn);
    m_info_ptr = png_create_info_struct(m_png_ptr);
}

CPngSaver::~CPngSaver() {
    if (m_png_ptr != NULL && m_info_ptr != NULL) {
        png_destroy_write_struct(&m_png_ptr, &m_info_ptr);
    } else if (m_png_ptr != NULL) {
        png_destroy_write_struct(&m_png_ptr, NULL);
    }
}

CPngSaver& CPngSaver::SetCompressLevel(int nCompressLevel) {
    Themis_DebugAssert((Z_NO_COMPRESSION <= nCompressLevel && nCompressLevel <= Z_BEST_COMPRESSION) ||
        nCompressLevel == Z_DEFAULT_COMPRESSION);
    png_set_compression_level(m_png_ptr, nCompressLevel);
    return *this;
}

/*
# png_set_filter()֐png_set_compression_level()֐ǉ@itB^шkx̐ݒj
# png_set_gAMA()֐ǉ@it@CEK}̐ݒj
# png_set_tIME()֐ǉ@iŏIC̐ݒj
# png_set_text()֐ǉ@i^CgCҁC}̐C쐬C\tgEFA̐ݒj
png_set_compression_mem_level
*/

void CPngSaver::Save(const CFileName& oFileName, const CPlane& src, const CRect& rc) {
    Save(CFileWriteStream(oFileName), src, rc);
}

namespace {
    void PngWriteFunction(png_structp png_ptr, png_bytep data, png_size_t length) {
        IWriteStream& rStream = *static_cast<IWriteStream*>(png_get_io_ptr(png_ptr));
        rStream.Write(data, length);
    }
    void PngFlushFunction(png_structp png_ptr) {
    }
}
void CPngSaver::Save(IWriteStream& rStream, const CPlane& src, const CRect& rc) {
    if (src.GetBpp() == 24) {
        //C++ȃIuWFNgsetjmp̑Oɒ`Ƃ
        boost::scoped_array<png_bytep> apPngRow;
        //Cȍ\O...
        if (setjmp(m_png_ptr->jmpbuf)) {
            throw ImageSaveException();
        }
        // pngIuWFNg̏
     // png_init_io(m_png_ptr, fp);
        png_set_write_fn(m_png_ptr, &rStream, &PngWriteFunction, &PngFlushFunction);
        png_set_filter(m_png_ptr, PNG_FILTER_TYPE_BASE, PNG_FILTER_SUB);
        png_set_IHDR(m_png_ptr, m_info_ptr, rc.GetWidth(), rc.GetHeight(), 8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
        png_write_info(m_png_ptr, m_info_ptr);
        // v[ւ̃|C^ݒ
        apPngRow.reset(new png_bytep[rc.GetHeight()]);
        Themis_PlaneLock1(src, rc.GetPoint(), p, nPitch);
        for (int y = 0, cy = rc.GetHeight(); y < cy; y++) {
            apPngRow[y] = static_cast<png_bytep>(p);
            p += nPitch;
        }
        // o
        png_set_bgr(m_png_ptr);
        png_write_image(m_png_ptr, apPngRow.get());
        png_write_end(m_png_ptr, m_info_ptr);
    } else if (src.GetBpp() == 8) {
        /*
            24̂RsyāAPNG_COLOR_TYPE_RGB  PNG_COLOR_TYPE_GRAYƁA
            png_set_bgr(m_png_ptr); B
        */
        //C++ȃIuWFNgsetjmp̑Oɒ`Ƃ
        boost::scoped_array<png_bytep> apPngRow;
        //Cȍ\O...
        if (setjmp(m_png_ptr->jmpbuf)) {
            throw ImageSaveException();
        }
        // pngIuWFNg̏
     // png_init_io(m_png_ptr, fp);
        png_set_write_fn(m_png_ptr, &rStream, &PngWriteFunction, &PngFlushFunction);
        png_set_filter(m_png_ptr, PNG_FILTER_TYPE_BASE, PNG_FILTER_SUB);
        png_set_IHDR(m_png_ptr, m_info_ptr, rc.GetWidth(), rc.GetHeight(), 8, PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
        png_write_info(m_png_ptr, m_info_ptr);
        // v[ւ̃|C^ݒ
        apPngRow.reset(new png_bytep[rc.GetHeight()]);
        Themis_PlaneLock1(src, rc.GetPoint(), p, nPitch);
        for (int y=0, cy=rc.GetHeight(); y<cy; y++) {
            apPngRow[y] = static_cast<png_bytep>(p);
            p += nPitch;
        }
        // o
        png_write_image(m_png_ptr, apPngRow.get());
        png_write_end(m_png_ptr, m_info_ptr);
    } else {
        CMemoryPlane oTmpPlane(24);
        oTmpPlane.SetSize(rc.GetSize());
        oTmpPlane.Blt(src, CPoint(), rc);
        Save(rStream, oTmpPlane, rc.GetSize().ToRect());
    }
}

#endif//Themis_ConfigPngLib

}
