#include "stdafx.h"
#include "Tga.h"
#include "../File.h"
#include "../Graphics.h"

#define Themis_TargaType 1

#include "../Utility/memio.h"
#define FILE Themis::MFILE
#define fopen Themis::fopen
using Themis::fread;
using Themis::fseek;
using Themis::fwrite;
#if Themis_TargaType == 0
    #include "../../../external/libtarga/libtarga.c"
#else
    namespace Themis { namespace Detail {
        static inline uint8_t* realloc(uint8_t* memory, size_t newSize) {
            return (uint8_t*)::realloc(memory, newSize);
        }
    }}
    using Themis::Detail::realloc;
    #include "../../../external/targa/targa.c"
#endif
#undef FILE
#undef fopen

namespace Themis {

CTgaLoader::CTgaLoader() {
}

CTgaLoader::~CTgaLoader() {
}

bool CTgaLoader::Load(const std::tr1::shared_ptr<CFileData>& fileData, bool bRenderOnlyAtOnce) {
    try {
        size_t fileSize = fileData->GetMemory().size();
        size_t* data[2] = {reinterpret_cast<size_t*>(&fileData->GetMemory()[0]), &fileSize};

    #if Themis_TargaType == 0
        int width, height;
        uint32* tga = static_cast<uint32*>(tga_load(
            reinterpret_cast<const char*>(data),
            &width, &height, TGA_TRUECOLOR_32));
        if (tga == NULL) {
            return false;
        }

        plane.reset(new CMemoryPlane(32));
        plane->SetSize(CSize(width, height));

        // RGBA  BGRAɕϊ㉺]Rs
        Themis_PlaneLock4(*plane, CPoint(0, height - 1), p, nPitch);
        Themis_PlaneGetSize(plane->GetSize(), nSrcW, nSrcH);
        for (int y = 0; y < nSrcH; y++) {
            for(int x = 0; x < nSrcW; x++) {
                Themis_PlaneGetBGRA32(tga[x], a, b, g, r);
                p[x] = Detail::BGRA32(a, r, g, b);
            }
            p -= nPitch;
            tga += width;
        }
  #else
        tga_image tga;
        tga_result result = tga_read(&tga,
            reinterpret_cast<const char*>(data));
        if (result != TGA_NOERR) {
            return false;
        }
        if (tga.width <= 0 || tga.height <= 0) {
            tga_free_buffers(&tga);
            return false;
        }
        if (tga_is_colormapped(&tga)) tga_color_unmap(&tga);
        if (tga_is_right_to_left(&tga)) tga_flip_horiz(&tga);
        //if (!tga_is_top_to_bottom(&tga)) tga_flip_vert(&tga);

        plane.reset(new CMemoryPlane(32));
        plane->SetSize(CSize(tga.width, tga.height));

        Themis_PlaneLock4(*plane, CPoint(), p, nPitch);
        Themis_PlaneGetSize(plane->GetSize(), nSrcW, nSrcH);
        const int pixelDepth = tga.pixel_depth;
        for (int y = 0; y < nSrcH; y++) {
            uint8_t* src = tga_find_pixel(&tga, 0, y);
            for (int x = 0; x < nSrcW; x++) {
                uint8_t b, g, r, a;
                tga_unpack_pixel(src + x * pixelDepth / 8,
                    pixelDepth, &b, &g, &r, &a);
                p[x] = Detail::BGRA32(0xff - a, r, g, b);
            }
            p += nPitch;
        }

        tga_free_buffers(&tga);
  #endif

        return true;
    } catch (...) { // } catch (std::exception&) {
        return false;
    }
}

CSize CTgaLoader::GetSize() const {
    return plane->GetSize();
}

void CTgaLoader::Render(CPlane& dst, const CPoint& pt, bool bRenderOnlyAtOnce) {
    dst.Blt(*plane, pt, plane->GetSize().ToRect());
}

}
