#include "stdafx.h"
#include "Layout.h"
namespace Themis {

void ILayout::Update() const {
    if (!GetScreen().IsValid() || !GetDocument().IsValid()) {
        dst = CRect();
        src = CRect();
    } else {
        CRect virtualDst = CalculateVirtualDst();
        CRect dst = virtualDst;
        CRect src = GetDocument();

        // ͂ݏo␳
        if (dst.IsValid() && src.IsValid()) {
            //double rateH = GetRateH();
            //double rateV = GetRateV();
            double rateH = static_cast<double>(virtualDst.GetWidth()) / GetDocument().GetWidth();
            double rateV = static_cast<double>(virtualDst.GetHeight()) / GetDocument().GetHeight();
            //]
            CRect rcDstOld = dst;
            dst.SetIntersect(GetScreen());
            CSize szDstGap1 = rcDstOld.GetPoint() - dst.GetPoint();   //Ə
            CSize szDstGap2 = CSize(rcDstOld.right - dst.right, rcDstOld.bottom - dst.bottom);  //EƉ
            src.left -= static_cast<int>(floor(szDstGap1.cx / rateH + 0.5));
            src.top -= static_cast<int>(floor(szDstGap1.cy / rateV + 0.5));
            src.right -= static_cast<int>(floor(szDstGap2.cx / rateH + 0.5));
            src.bottom -= static_cast<int>(floor(szDstGap2.cy / rateV + 0.5));
            //]
            CRect rcSrcOld = src;
            src.SetIntersect(GetDocument());
            CSize szSrcGap1 = rcSrcOld.GetPoint() - src.GetPoint();   //Ə
            CSize szSrcGap2 = CSize(rcSrcOld.right - src.right, rcSrcOld.bottom - src.bottom);  //EƉ
            dst.left -= static_cast<int>(floor(szSrcGap1.cx * rateH + 0.5));
            dst.top -= static_cast<int>(floor(szSrcGap1.cy * rateV + 0.5));
            dst.right -= static_cast<int>(floor(szSrcGap2.cx * rateH + 0.5));
            dst.bottom -= static_cast<int>(floor(szSrcGap2.cy * rateV + 0.5));

         // Themis_DebugAssert(fabs(static_cast<double>(dst.GetWidth()) / src.GetWidth() - rateH) <= 0.1);
         // Themis_DebugAssert(fabs(static_cast<double>(dst.GetHeight()) / src.GetHeight() - rateV) <= 0.1);
        }
        
        this->src = src;
        this->dst = dst;
    }
}

/// DocumentɑΉ]W̎Zo
CRect SimpleLayout::CalculateVirtualDst() const {
    // documentViewArea  Screen Ƀ}bsO銴ŁA
    // Document }bsOWԂB
    CRect screen = GetScreen();
    double rateH = static_cast<double>(screen.GetWidth()) / documentViewArea.GetWidth();
    double rateV = static_cast<double>(screen.GetHeight()) / documentViewArea.GetHeight();
    CRect document = GetDocument();
    CRect virtualDst;
    virtualDst.left = screen.left + static_cast<int>(floor((document.left - documentViewArea.left) * rateH + 0.5));
    virtualDst.top = screen.top + static_cast<int>(floor((document.top - documentViewArea.top) * rateV + 0.5));
    virtualDst.right = screen.right + static_cast<int>(floor((document.right - documentViewArea.right) * rateH + 0.5));
    virtualDst.bottom = screen.bottom + static_cast<int>(floor((document.bottom - documentViewArea.bottom) * rateV + 0.5));
    return virtualDst;
}

/// ̍쐬
std::tr1::shared_ptr<ILayout> SimpleLayout::Clone() const {
    std::tr1::shared_ptr<SimpleLayout> p(new SimpleLayout());
    p->SetScreen(GetScreen());
    p->SetDocument(GetDocument());
    p->SetDocumentViewArea(GetDocumentViewArea());
    return p;
}

/// DocumentɑΉ]W̎Zo
CRect StretchingLayoutBase::CalculateVirtualDst() const {
    // TCY͔hNXŁB
    const CSize size = CalculateVirtualDstSize();
    // zu
    const CRect screen = GetScreen();
    CPoint pos;
    switch (alignH) {
    case 0: pos.x = screen.left; break;
    case 1: pos.x = screen.left + (screen.GetWidth() - size.cx) / 2; break;
    case 2: pos.x = screen.right - size.cx; break;
    default: Themis_DebugWarning(L"StretchingLayoutBasealignHs"); break;
    }
    switch (alignV) {
    case 0: pos.y = screen.top; break;
    case 1: pos.y = screen.top + (screen.GetHeight() - size.cy) / 2; break;
    case 2: pos.y = screen.bottom - size.cy; break;
    default: Themis_DebugWarning(L"StretchingLayoutBasealignVs"); break;
    }
    return CRect(pos, size);
}

/// DocumentɑΉ]TCY̎Zo
CSize AutoStretchingLayout::CalculateVirtualDstSize() const {
    CSize screenSize = GetScreen().GetSize();
    CSize documentSize = GetDocument().GetSize();
    double dRateX = static_cast<double>(screenSize.cx) / documentSize.cx;
    double dRateY = static_cast<double>(screenSize.cy) / documentSize.cy;
    if (m_bStickOut) {
        stretchToHorz = dRateY < dRateX;
    } else {
        stretchToHorz = dRateX < dRateY;
    }
    double dRate = stretchToHorz ? dRateX : dRateY;
    if ((dRate < 1.0 && m_bReduction) || (1.0 < dRate && m_bMagnification)) {
        // k܂͊g
        return CSize(static_cast<int>(floor(documentSize.cx * dRate + 0.5)),
                     static_cast<int>(floor(documentSize.cy * dRate + 0.5)));
    } else {
        // {
        return documentSize;
    }
}

/// ̍쐬
std::tr1::shared_ptr<ILayout> AutoStretchingLayout::Clone() const {
    std::tr1::shared_ptr<AutoStretchingLayout> p(new AutoStretchingLayout());
    p->SetScreen(GetScreen());
    p->SetDocument(GetDocument());
    p->SetAlignV(GetAlignV());
    p->SetAlignH(GetAlignH());
    p->SetReduction(IsReduction());
    p->SetMagnification(IsMagnification());
    p->SetStickOut(IsStickOut());
    return p;
}

/// DocumentɑΉ]TCY̎Zo
CSize FixedStretchingLayout::CalculateVirtualDstSize() const {
    CSize documentSize = GetDocument().GetSize();
    return CSize(static_cast<int>(floor(documentSize.cx * rate + 0.5)),
                 static_cast<int>(floor(documentSize.cy * rate + 0.5)));
}

/// ̍쐬
std::tr1::shared_ptr<ILayout> FixedStretchingLayout::Clone() const {
    std::tr1::shared_ptr<FixedStretchingLayout> p(new FixedStretchingLayout());
    p->SetScreen(GetScreen());
    p->SetDocument(GetDocument());
    p->SetAlignV(GetAlignV());
    p->SetAlignH(GetAlignH());
    p->SetRate(GetRate());
    return p;
}

}
