#include "stdafx.h"
#include <time.h>
#include "Windows/System.h"
#include "Thread/Thread.h"
#include "Thread/ThreadPool.h"
#include "Archive/Zip.h"
#include "Config.h"
#include "TempFile.h"
#include "File.h"
#include "Http.h"
#include "BatchFile.h"
#include "MD5.h"
#include "NetworkUpdaterBase.h"
namespace Themis {

CNetworkUpdateInfo::SFile::SFile() : nSize(0) {
}

string CNetworkUpdateInfo::SFile::ToStr() const {
    return strName;
}

CNetworkUpdateInfo::SEntry::SEntry() : strDst(L"./"), bBackup(true), bDelete(false) {
}

string CNetworkUpdateInfo::SEntry::ToStr() const {
    return strSrc;
}

CNetworkUpdateInfo::SNotice::SNotice() {
    time_t tt = time(NULL);
    tm* t = localtime(&tt);
    strData = string::Format(L"[%d/%02d/%02d]\r\n", t->tm_year + 1900, t->tm_mon + 1, t->tm_mday);
}

string CNetworkUpdateInfo::SNotice::ToStr() const {
    return strData;
}

//------------------------------------------------------------------//

CNetworkUpdateInfo::CNetworkUpdateInfo() {
}

CNetworkUpdateInfo::~CNetworkUpdateInfo() {
}

// File

void CNetworkUpdateInfo::AddFile(const SFile& oFile) {
    m_oFileData.push_back(oFile);
}

void CNetworkUpdateInfo::DelFile(int i) {
    if (0 <= i && i < GetFileCount()) {
        m_oFileData.erase(m_oFileData.begin() + i);
    }
}

bool CNetworkUpdateInfo::UpFile(int i) {
    if (0 < i && i < GetFileCount()) {
        std::swap(m_oFileData[i], m_oFileData[i - 1]);
        return true;
    }
    return false;
}

bool CNetworkUpdateInfo::DownFile(int i) {
    if (0 <= i && i + 1 < GetFileCount()) {
        std::swap(m_oFileData[i], m_oFileData[i + 1]);
        return true;
    }
    return false;
}

int CNetworkUpdateInfo::GetFileCount() const {
    return static_cast<int>(m_oFileData.size());
}

CNetworkUpdateInfo::SFile& CNetworkUpdateInfo::GetFile(int i) {
    return m_oFileData[i];
}

const CNetworkUpdateInfo::SFile& CNetworkUpdateInfo::GetFile(int i) const {
    return m_oFileData[i];
}

// Entry

void CNetworkUpdateInfo::AddEntry(const SEntry& oEntry) {
    m_oEntryData.push_back(oEntry);
}

void CNetworkUpdateInfo::DelEntry(int i) {
    if (0 <= i && i < GetEntryCount()) {
        m_oEntryData.erase(m_oEntryData.begin() + i);
    }
}

bool CNetworkUpdateInfo::UpEntry(int i) {
    if (0 < i && i < GetEntryCount()) {
        std::swap(m_oEntryData[i], m_oEntryData[i - 1]);
        return true;
    }
    return false;
}

bool CNetworkUpdateInfo::DownEntry(int i) {
    if (0 <= i && i + 1 < GetEntryCount()) {
        std::swap(m_oEntryData[i], m_oEntryData[i + 1]);
        return true;
    }
    return false;
}

int CNetworkUpdateInfo::GetEntryCount() const {
    return static_cast<int>(m_oEntryData.size());
}

CNetworkUpdateInfo::SEntry& CNetworkUpdateInfo::GetEntry(int i) {
    return m_oEntryData[i];
}

const CNetworkUpdateInfo::SEntry& CNetworkUpdateInfo::GetEntry(int i) const {
    return m_oEntryData[i];
}

// Notice

void CNetworkUpdateInfo::AddNotice(const SNotice& oNotice) {
    m_oNoticeData.push_back(oNotice);
}

void CNetworkUpdateInfo::DelNotice(int i) {
    if (0 <= i && i < GetNoticeCount()) {
        m_oNoticeData.erase(m_oNoticeData.begin() + i);
    }
}

bool CNetworkUpdateInfo::UpNotice(int i) {
    if (0 < i && i < GetNoticeCount()) {
        std::swap(m_oNoticeData[i], m_oNoticeData[i - 1]);
        return true;
    }
    return false;
}

bool CNetworkUpdateInfo::DownNotice(int i) {
    if (0 <= i && i + 1 < GetNoticeCount()) {
        std::swap(m_oNoticeData[i], m_oNoticeData[i + 1]);
        return true;
    }
    return false;
}

int CNetworkUpdateInfo::GetNoticeCount() const {
    return static_cast<int>(m_oNoticeData.size());
}

CNetworkUpdateInfo::SNotice& CNetworkUpdateInfo::GetNotice(int i) {
    return m_oNoticeData[i];
}

const CNetworkUpdateInfo::SNotice& CNetworkUpdateInfo::GetNotice(int i) const {
    return m_oNoticeData[i];
}

/// f[^̓ǂݍ
void CNetworkUpdateInfo::Load(const CFileName& oFileName) {
    CFile oFile(oFileName);
    if (oFile.Exists()) {
        LoadFromStr(oFile.LoadStr());
    } else {
        LoadFromStr(string());
    }
}
void CNetworkUpdateInfo::LoadFromStr(const string& strData) {
    CConfig oConfig;
    oConfig.Num(L"File", 0);
    oConfig.Num(L"Entry", 0);
    oConfig.Num(L"Notice", 0);
    oConfig.LoadFromStr(strData);
    m_oFileData.resize(oConfig.Num(L"File"));
    for (size_t i = 0, n = m_oFileData.size(); i < n; i++) {
        m_oFileData[i].strName = oConfig.File(string::Format(L"File%03d", i));
        m_oFileData[i].strTime = oConfig.Indirect(string::Format(L"File%03dTime", i));
        m_oFileData[i].strMD5 = oConfig.Indirect(string::Format(L"File%03dMD5", i));
        m_oFileData[i].nSize = oConfig.Num(string::Format(L"File%03dSize", i));
    }
    m_oEntryData.resize(oConfig.Num(L"Entry"));
    for (size_t i = 0, n = m_oEntryData.size(); i < n; i++) {
        m_oEntryData[i].strSrc = oConfig.File(string::Format(L"Entry%03dSrc", i));
        m_oEntryData[i].strDst = oConfig.File(string::Format(L"Entry%03dDst", i));
        m_oEntryData[i].bBackup = oConfig.Bool(string::Format(L"Entry%03dBackup", i));
        m_oEntryData[i].bDelete = oConfig.Bool(string::Format(L"Entry%03dDelete", i));
    }
    m_oNoticeData.resize(oConfig.Num(L"Notice"));
    for (size_t i = 0, n = m_oNoticeData.size(); i < n; i++) {
        m_oNoticeData[i].strData = oConfig.Str(string::Format(L"Notice%03d", i));
    }
}

void CNetworkUpdateInfo::UpdateFileTimeSize() {
    for (size_t i = 0, n = m_oFileData.size(); i < n; i++) {
        CFile oFile(CFileNameSystemDir(m_oFileData[i].strName));
        FILETIME ft = oFile.GetLastUpdate();
        SYSTEMTIME st = {0};
        ::FileTimeToSystemTime(&ft, &st);
        m_oFileData[i].strTime = string::Format(L"%d/%02d/%02d %02d:%02d:%02d", st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond);
        m_oFileData[i].strMD5 = CMD5::Make(oFile.LoadBin());
        m_oFileData[i].nSize = static_cast<int>(oFile.GetSize());
    }
}

void CNetworkUpdateInfo::Save(const CFileName& oFileName) {
    CConfig oConfig;
    oConfig.Num(L"File", static_cast<int>(m_oFileData.size()));
    for (size_t i = 0, n = m_oFileData.size(); i < n; i++) {
        oConfig.File(string::Format(L"File%03d", i), m_oFileData[i].strName);
        oConfig.Indirect(string::Format(L"File%03dTime", i), m_oFileData[i].strTime);
        oConfig.Indirect(string::Format(L"File%03dMD5", i), m_oFileData[i].strMD5);
        oConfig.Num(string::Format(L"File%03dSize", i), m_oFileData[i].nSize);
    }
    oConfig.Num(L"Entry", static_cast<int>(m_oEntryData.size()));
    for (size_t i = 0, n = m_oEntryData.size(); i < n; i++) {
        oConfig.File(string::Format(L"Entry%03dSrc", i), m_oEntryData[i].strSrc);
        oConfig.File(string::Format(L"Entry%03dDst", i), m_oEntryData[i].strDst);
        oConfig.Bool(string::Format(L"Entry%03dBackup", i), m_oEntryData[i].bBackup);
        oConfig.Bool(string::Format(L"Entry%03dDelete", i), m_oEntryData[i].bDelete);
    }
    oConfig.Num(L"Notice", static_cast<int>(m_oNoticeData.size()));
    for (size_t i = 0, n = m_oNoticeData.size(); i < n; i++) {
        // sR[hCR+LFɂẢs𑵂ĂB
        {
            vector<string> vct = m_oNoticeData[i].strData.TrimEnd().Split(Array<string>(L"\r\n", L"\n", L"\r"));
            m_oNoticeData[i].strData = string::Join(L"\r\n", vct);
            m_oNoticeData[i].strData += L"\r\n";
        }
        oConfig.Str(string::Format(L"Notice%03d", i), m_oNoticeData[i].strData);
    }
    string strData = oConfig.SaveToStr();
    if (CFile(oFileName).LoadStr() != strData) { // Ăꍇ̂ݕۑ
        CFile(oFileName).Save(strData);
    }
}

/// t@CTCY̍v擾
int CNetworkUpdateInfo::GetFileSizeSum() const {
    int nSize = 0;
    for (size_t i = 0, n = m_oFileData.size(); i < n; i++) {
        nSize += m_oFileData[i].nSize;
    }
    return nSize;
}

//------------------------------------------------------------------//

CNetworkUpdaterBase::CNetworkUpdaterBase() {
    threadValid = false;
}

CNetworkUpdaterBase::~CNetworkUpdaterBase() {
    StopUpdate();
}

/// WM_INITDIALOGƂĂԂׂ
void CNetworkUpdaterBase::UpdateGetInfo() {
    Themis_DebugAssert(!threadValid);
    m_nMode = 0;
    threadValid = true;
    thread = ThreadPool::CreateWorkerThread();
    thread->SetWork(std::tr1::bind(&CNetworkUpdaterBase::OnGetInfoThread, this));
}

/// GetInfo()Iđs{^ꂽ肵ĂԂׂ
void CNetworkUpdaterBase::Update() {
    Themis_DebugAssert(!threadValid);
    m_nMode = 1;
    threadValid = true;
    thread = ThreadPool::CreateWorkerThread();
    thread->SetWork(std::tr1::bind(&CNetworkUpdaterBase::OnUpdateThread, this));
}

/// 𒆒f(Ԃ)
void CNetworkUpdaterBase::CancelUpdate() {
    threadValid = false;
}

/// 𒆒f(I܂ő҂)
void CNetworkUpdaterBase::StopUpdate() {
    if (threadValid) {
        threadValid = false;
        thread->Join();
    }
}

/// GetInfo()ȏ
void CNetworkUpdaterBase::OnGetInfoThread() {
    Thread::SetName(L"NetworkUpdaterGetInfoThread");
    try {
        // hNXFX擾
        CConfig& rConfig = GetConfig();
        const string strBaseName = GetConfigBaseName();
        const string strUpdateUrl = GetUpdateUrl();
        // Âm̕\
        {
            vector<string> vctOldNotice(rConfig.Num(strBaseName + L"OldNotice"));
            for (size_t i = 0, n = vctOldNotice.size(); i < n; i++) {
                vctOldNotice[i] = rConfig.Str(strBaseName + L"OldNotice" + string::Format(L"%03d", i));
            }
            // fĂ߂[
            if (!threadValid) {
                return;
            }
            OnOldNotice(vctOldNotice);
        }
        // GETNGXg
        {
            SHttpReceiveData oData;
            HttpGet(strUpdateUrl, oData);
            // fĂ߂[
            if (!threadValid) {
                return;
            }
            // XVǂݍ݁[
            m_oInfo.LoadFromStr(oData.GetDataAsStr());
        }
        // XVt@C邩ǂ`FbN
        if (!ExistsUpdateFile()) {
            // Abvf[gt@ĈŔhNXɒʒmētL^ďI
         // rConfig.Indirect(strBaseName + L"Update", m_strLastModified);
            OnGetInfoNoUpdate();
            return;
        }
        // VmRtBOɕۑ
        {
            rConfig.Num(strBaseName + L"OldNotice", m_oInfo.GetNoticeCount());
            for (int i = 0, n = m_oInfo.GetNoticeCount(); i < n; i++) {
                rConfig.Str(strBaseName + L"OldNotice" + string::Format(L"%03d", i), m_oInfo.GetNotice(i).ToStr());
            }
        }
        // GetInfo()Iʒm
        OnGetInfo(m_oInfo);
        threadValid = false;
    } catch (std::exception&) {
        OnError();
        threadValid = false;
        throw;
    }
}

/// Update()ȏ
void CNetworkUpdaterBase::OnUpdateThread() {
    Thread::SetName(L"NetworkUpdaterUpdateThread");
    try {
        // hNXFX擾
        CConfig& rConfig = GetConfig();
        const string strBaseName = GetConfigBaseName();
        const string strUpdateUrl = GetUpdateUrl();
        const string strBaseNameFile = strBaseName + L"File";
        // GETNGXgȌ
        m_nCurrentOld = 0;
        m_nCurrentAll = 0;
        m_nMaxAll = m_oInfo.GetFileSizeSum();
        // e|tH_̍쐬
        m_pTempDir.reset(new CTempDir(CFileNameModuleDir()));
        m_pTempDir->Create();
        // XVt@C_E[h
        for (int i = 0, n = m_oInfo.GetFileCount(); i < n; i++) {
            // fĂ߂[
            if (!threadValid) {
                return;
            }
            const string str = strBaseNameFile + string::Format(L"%03d", i);
            if (!rConfig.Exists(str) || rConfig.Indirect(str) != m_oInfo.GetFile(i).strTime) {
                m_strCurrentFile = m_oInfo.GetFile(i).strName;
                SHttpReceiveData oData;
                HttpGet(CFileName(strUpdateUrl).GetDir() + m_strCurrentFile, oData);
                if (m_oInfo.GetFile(i).strMD5 != CMD5::Make(oData.vctData)) {
                    throw RuntimeException(L"t@C '" + m_strCurrentFile + L"' jĂۂł");
                }
                // fĂ߂[
                if (!threadValid) {
                    return;
                }
                // _E[ht@Cۑ
                if (CFileName(m_strCurrentFile).GetSuffixLc() == L"zip") {
                    // zipȏꍇ𓀂B璼ڂ͍ĂȂ̂ŁA
                    // _Tǈxt@CɏłBB
                    CTempFile oTempFile(m_pTempDir->GetName(), string(), L"zip");
                    oTempFile.Create(oData.vctData);
                    CZipArchiver oArchiver;
                    oArchiver.Open(oTempFile.GetName());
                    // utemp/`.zip/ɓt@CvƂ
                    CFileName oZipDirName(m_strCurrentFile + L"\\", m_pTempDir->GetName());
                    for (int i = 0, n = oArchiver.GetFileCount(); i < n; i++) {
                        oArchiver.ExtractFile(i, oZipDirName);
                    }
                } else {
                    // ʏt@C̏ꍇ̂܂ܕۑ
                    CFile(CFileName(m_strCurrentFile, m_pTempDir->GetName())).Save(oData.vctData);
                }
            }
        }
        // fĂ߂[
        if (!threadValid) {
            return;
        }
        // ŏImF
        if (!ConfirmUpdate()) {
            return; // LZꂽ
        }
        // obNAbv
        try { 
            if (rConfig.Bool(strBaseName + L"Backup")) {
                int nBackupCount = rConfig.Num(strBaseName + L"BackupCount");
                for (int i = nBackupCount - 1; ; i++) { // BackupCountȏ̂S폜
                    CFile oDir(CFileNameModuleDir(string::Format(L"$backup-%02d$\\", i)));
                    if (oDir.Exists()) {
                        oDir.Delete(true);
                    } else {
                        break;
                    }
                }
                for (int i = nBackupCount - 2; 0 <= i; i--) { // 0 ` BackupCount-2܂łl[
                    CFile oDir(CFileNameModuleDir(string::Format(L"$backup-%02d$\\", i)));
                    if (oDir.Exists()) {
                        oDir.Move(CFileNameModuleDir(string::Format(L"$backup-%02d$\\", i + 1)));
                    }
                }
                for (int i = 0, n = m_oInfo.GetEntryCount(); i < n; i++) { // ݂̂0փRs[
                    CFileNameModuleDir oDir(m_oInfo.GetEntry(i).strDst);
                    string strName = CFileName(m_oInfo.GetEntry(i).strSrc).GetPureName();
                    CFile oFile(CFileName(strName, oDir));
                    if (oFile.Exists()) {
                        if (m_oInfo.GetEntry(i).bDelete) {
                            // ƂDLLƍ폜ȂɂȂĂ܂A
                            // ob`t@Cɓ̂ʓ|Ȃ̂łƂ肠蔲
                            oFile.Delete();
                        } else {
                            // Rs邾
                            oFile.Copy(CFileName(m_oInfo.GetEntry(i).strDst + strName, CFileNameModuleDir(L"$backup-00$\\")));
                        }
                    }
                }
                OnEndBackup();
            }
        } catch (...) {
            // ߂ǂ疳BȂǂȂ񂾂ǁBB(L`)
        }
        // ob`t@C쐬
        CFileName moveBatchName(L"$movefile$.bat", m_pTempDir->GetName());
        CFileName updateBatchName(L"$update$.bat", m_pTempDir->GetName());
        {
            CBatchFile oMoveBatch;
            oMoveBatch.Add(L"mkdir %2");
            oMoveBatch.Add(L":retry");
            oMoveBatch.Add(L"move /Y %1 %2");
            oMoveBatch.Add(L"if exist %1 goto retry");
        #ifndef NDEBUG
            oMoveBatch.Add(L"pause");
        #endif
            oMoveBatch.Save(moveBatchName);
        }
        {
            CBatchFile oUpdateBatch;
            for (int i = 0, n = m_oInfo.GetEntryCount(); i < n; i++) {
                const string& strSrc = CFileName::SlashToYen(m_oInfo.GetEntry(i).strSrc);
                string strDst = CFileNameModuleDir(m_oInfo.GetEntry(i).strDst).RemoveDirSeparator();
                oUpdateBatch.Add(L"call", CBatchFile::AddWQuote(moveBatchName.Get()),
                    CBatchFile::AddWQuote(CFileName(strSrc, m_pTempDir->GetName()).Get()),
                    CBatchFile::AddWQuote(CFileNameModuleDir(strDst).Get()));
            }
            oUpdateBatch.Add(L"cd ..");
            oUpdateBatch.Add(L"start", L"\"\"", CBatchFile::AddWQuote(CSystem::GetModuleFileName()));
            oUpdateBatch.Add(L"rmdir", L"/S /Q", CBatchFile::AddWQuote(m_pTempDir->GetName().Get()));
        #ifndef NDEBUG
            oUpdateBatch.Add(L"pause");
        #endif
            oUpdateBatch.Run(updateBatchName);
        }
        // I
        m_pTempDir->Release();
        for (int i = 0, n = m_oInfo.GetFileCount(); i < n; i++) { // Vt@CMD5L^Ă
            rConfig.Indirect(strBaseNameFile + string::Format(L"%03dMD5", i), m_oInfo.GetFile(i).strMD5);
        }
     // rConfig.Indirect(strBaseName + L"Update", m_strLastModified);
        rConfig.Delete(strBaseName + L"Update");
        PostQuit();
        threadValid = false;
    } catch (...) {
        OnError();
        threadValid = false;
    }
}

void CNetworkUpdaterBase::HttpHead(const string& strUrl, SHttpReceiveData& rData) {
    // hNXFX擾
    CConfig& rConfig = GetConfig();
    const string strBaseName = GetConfigBaseName();
    // HTTPwb_擾
    CHttp oHttp;
    if (rConfig.Bool(strBaseName + L"Proxy")) {
        oHttp.SetProxy(rConfig.Indirect(strBaseName + L"ProxyServer"));
    }
    oHttp.SetNoCache(true); // LbVʂɒƂ΍H
    oHttp.Head(strUrl);
    oHttp.ReceiveData(rData);
}

void CNetworkUpdaterBase::HttpGet(const string& strUrl, SHttpReceiveData& rData) {
    // hNXFX擾
    CConfig& rConfig = GetConfig();
    const string strBaseName = GetConfigBaseName();
    // HTTPf[^擾
    CHttp oHttp;
    if (rConfig.Bool(strBaseName + L"Proxy")) {
        oHttp.SetProxy(rConfig.Indirect(strBaseName + L"ProxyServer"));
    }
    oHttp.SetNoCache(true); // LbVʂɒƂ΍H
    oHttp.Get(strUrl);
    oHttp.ReceiveData(rData, *this);
}

/// Abvf[gt@C邩ǂ
bool CNetworkUpdaterBase::ExistsUpdateFile() {
    // hNXFX擾
    CConfig& rConfig = GetConfig();
    const string strBaseName = GetConfigBaseName();
    const string strBaseNameFile = strBaseName + L"File";
    // [v񂵂đSt@C`FbN
    for (int i = 0, n = m_oInfo.GetFileCount(); i < n; i++) {
        const string str = strBaseNameFile + string::Format(L"%03dMD5", i);
        if (!rConfig.Exists(str) || rConfig.Indirect(str) != m_oInfo.GetFile(i).strMD5) {
            return true; // łႤ̂ true
        }
    }
    return false;
}

void CNetworkUpdaterBase::OnProgress(unsigned nCurrent, unsigned nMax) {
    if (m_nMode == 0) {
        OnGetInfoProgress(nCurrent, nMax);
    } else {
        m_nCurrentAll += nCurrent - m_nCurrentOld;
        m_nCurrentOld = nCurrent;
        OnUpdateProgress(m_strCurrentFile, nCurrent, nMax, m_nCurrentAll, m_nMaxAll);
    }
}


}
