#pragma once
#include "Factory.h"
#include "Utility/Algorithm.h"
namespace Themis {

/**
    IuWFNgv[̊ǗsB

    QƃJEgȂCX^X́A
    PoolɕԋpAȍ~CreateInstance()ōėpB
*/
template<class Object> class ObjectPool : public IFactory<Object> {
    boost::mutex mutex;
    vector<Object*> pooledObjects;
    int maxPoolCount; // őv[
  #ifndef NDEBUG
    int usingCount; // gp̃IuWFNg
  #endif
public:
    ObjectPool() : maxPoolCount(-1) {
      #ifndef NDEBUG
        usingCount = 0;
      #endif
    }
    ~ObjectPool() {
      #ifndef NDEBUG
        if (0 < usingCount) { // ȂȂ\ȂɂƂBB
            Themis_DebugLog(L"---------- ObjectPool --------------");
            Themis_DebugLog(string::Format(L"^fIuWFNgF%d", usingCount));
            Themis_DebugLog(string::Format(L"@v[IuWFNgF%d", pooledObjects.size()));
            Themis_DebugLog(L"-------------------------------------");
        }
      #endif
        for (size_t i = 0, n = pooledObjects.size(); i < n; i++) {
            boost::checked_delete(pooledObjects[i]);
        }
    }

    /// v[IuWFNg̍쐬
    std::tr1::shared_ptr<Object> CreateInstance() {
        Object* p;
        if (pooledObjects.empty()) {
            p = new Object;
        } else {
            boost::mutex::scoped_lock lock(mutex);
            if (pooledObjects.empty()) {
                p = new Object;
            } else {
                p = pooledObjects[pooledObjects.size() - 1];
                pooledObjects.pop_back();
            }
        }
      #ifndef NDEBUG
        ++usingCount;
      #endif
        return std::tr1::shared_ptr<Object>(p, std::tr1::bind(&ObjectPool<Object>::OnDelete, this, _1));
    }

    /// gpĂȂIuWFNg̍폜
    void Garbage() {
        vector<Object*> vct;
        {
            boost::mutex::scoped_lock lock(mutex);
            vct.swap(pooledObjects);
        }
        for (size_t i = 0, n = vct.size(); i < n; i++) {
            boost::checked_delete(vct[i]);
        }
    }

    /// v[ĂIuWFNǧ̎擾
    int GetCount() const {
        return static_cast<int>(pooledObjects.size());
    }
    /// v[ĂIuWFNgȂtrue
    bool IsEmpty() const { return GetCount() <= 0; }

    /// őv[̐ݒ
    void SetMaxPoolCount(int n) {
        maxPoolCount = n;
        if (maxPoolCount <= 0) {
            // v[[h
        } else if (maxPoolCount < GetCount()) {
            // v[[hB]͂̏ō폜
            vector<Object*> vct;
            {
                boost::mutex::scoped_lock lock(mutex);
                if (maxPoolCount < GetCount()) {
                    vct.resize(GetCount() - maxPoolCount);
                    std_copy(pooledObjects.begin() + maxPoolCount, pooledObjects.end(), vct.begin());
                    pooledObjects.resize(maxPoolCount);
                }
            }
            for (size_t i = 0, n = vct.size(); i < n; i++) {
                boost::checked_delete(vct[i]);
            }
        }
    }


private:
    void OnDelete(Object* p) {
        boost::mutex::scoped_lock lock(mutex);
        pooledObjects.push_back(p);
      #ifndef NDEBUG
        --usingCount;
      #endif
    }
};

}
