#include "stdafx.h"
#include "PageLoaderBase.h"
#include "Thread/Thread.h"
#include "Thread/ThreadPool.h"
namespace Themis {

CPageLoaderBase::CPageLoaderBase() {
    threadValid = true;
    thread = ThreadPool::CreateWorkerThread();
    thread->SetWork(std::tr1::bind(&CPageLoaderBase::OnThread, this));
}

CPageLoaderBase::~CPageLoaderBase() {
    Dispose();
}

void CPageLoaderBase::Dispose() {
    boost::recursive_mutex::scoped_lock lock(mutex);
    if (thread.get() != NULL) {
        threadValid = false;
        cacheCondition.notify_all();
        lock.unlock(); // ŃbNȂƃfbhbNB
        thread->Join();
        thread.reset();
    }
}

/// LbV̐ݒ
void CPageLoaderBase::SetCacheCount(int nCache, int nForward) {
    boost::recursive_mutex::scoped_lock lock(mutex);
    // assertƕ␳
    Themis_DebugAssert(1 <= nCache);
    Themis_DebugAssert(nForward <= nCache - 1);
    nCache = Max(nCache, 1);
    nForward = Min(nForward, nCache - 1);
    // LbVƂςĂFXB
    if (nCache != GetCacheCount() || nForward != GetForwardCount()) {
        int oldCacheCount = static_cast<int>(m_vctCaches.size());
        if (nCache < oldCacheCount) {
            m_vctCaches.erase(m_vctCaches.begin() + nCache, m_vctCaches.end());
        } else {
            for (int i = 0, n = nCache - oldCacheCount; i < n; i++) {
                m_vctCaches.push_back(new SCache(CreatePage()));
            }
        }
        m_nForwardCount = nForward;
        // xSjčēǂݍ
        m_nCurrent = 0; // LbVꍇ̂ł܂ŏɂ
        ReloadAllPage();
    }
}

/// ̃y[W
void CPageLoaderBase::ToNextPage() {
    Themis_DebugAssert(IsReady());
    boost::recursive_mutex::scoped_lock lock(mutex);
    // ̃LbVֈړĈLbVNAČ݂̂ǂݍ
    m_nCurrent = InnerGetRotatedIndex(m_nCurrent + 1);
    InnerClearPage(InnerGetRotatedIndex(m_nCurrent + GetForwardCount()));
    if (GetCache(m_nCurrent).state == eCacheStateEmpty) {
        InnerPrepareForwardPage(m_nCurrent);
        InnerLoadCurrent();
    }
    CurrentPageChanged();
}

/// Õy[W
void CPageLoaderBase::ToPriorPage() {
    Themis_DebugAssert(IsReady());
    boost::recursive_mutex::scoped_lock lock(mutex);
    // ÕLbVֈړĈLbVNAČ݂̂ǂݍ
    m_nCurrent = InnerGetRotatedIndex(m_nCurrent - 1);
    InnerClearPage(InnerGetRotatedIndex(m_nCurrent - GetBackwardCount()));
    if (GetCache(m_nCurrent).state == eCacheStateEmpty) {
        InnerPrepareBackwardPage(m_nCurrent);
        InnerLoadCurrent();
    }
    CurrentPageChanged();
}

/// ݂̃y[WȊOēǂݍ
void CPageLoaderBase::ReloadCachePage() {
    Themis_DebugAssert(IsReady());
    boost::recursive_mutex::scoped_lock lock(mutex);
    // ݂̃y[WȊOSĔj
    for (int i = 0, n = m_nCurrent; i < n; i++) {
        InnerClearPage(i);
    }
    for (int i = m_nCurrent + 1, n = GetCacheCount(); i < n; i++) {
        InnerClearPage(i);
    }
}

/// ReloadCurrentPage() + ReloadCachePage()
void CPageLoaderBase::ReloadAllPage() {
    Themis_DebugAssert(IsReady());
    boost::recursive_mutex::scoped_lock lock(mutex);
    // SĔj
    for (int i = 0, n = GetCacheCount(); i < n; i++) {
        InnerClearPage(i);
    }
    // ݂̂ǂݍ
    if (GetCache(m_nCurrent).state == eCacheStateEmpty) {
        InnerPrepareCurrentPage(m_nCurrent);
        InnerLoadCurrent();
    }
    CurrentPageChanged();
}

/// ݂̃y[WǂݍݒȂI܂ő҂
void CPageLoaderBase::WaitLoadPage() {
    Themis_DebugAssert(IsReady());
    boost::recursive_mutex::scoped_lock lock(mutex);
    Themis_DebugAssertLog(GetCache(m_nCurrent).state != eCacheStateEmpty);
 // ǂݍ݂Ɏs肵ĂĂ̂蓾邯ǁAꉞOƂ
    if (GetCache(m_nCurrent).state == eCacheStateLoading) { // ǂݍݒȂI܂ő҂B
        loadingCondition.wait(lock);
    }
    // ̎_œǂݍ݂͏IĂ͂B
    Themis_DebugAssert(GetCache(m_nCurrent).state == eCacheStateLoaded || GetCache(m_nCurrent).state == eCacheStateLoadFailed);
}

/// ǂݏJn
void CPageLoaderBase::StartLoadBeforehand() {
    boost::recursive_mutex::scoped_lock lock(mutex);
    cacheCondition.notify_all();
}

//------------------------------------------------------------------//

/// LbVf[^̔j
void CPageLoaderBase::InnerClearPage(int i) {
    GetCache(i).state = eCacheStateEmpty;
    ClearPage(*GetCache(i).pageData);
}

void CPageLoaderBase::InnerPrepareCurrentPage(int i) {
    Themis_DebugAssert(GetCache(i).state == eCacheStateEmpty);
    PrepareCurrentPage(*GetCache(i).pageData);
}

void CPageLoaderBase::InnerPrepareForwardPage(int i) {
    Themis_DebugAssert(GetCache(i).state == eCacheStateEmpty);
    int nPrior = InnerGetRotatedIndex(i - 1);
    Themis_DebugAssert(GetCache(nPrior).state != eCacheStateLoading);
    PrepareForwardPage(*GetCache(i).pageData, *GetCache(nPrior).pageData);
}

void CPageLoaderBase::InnerPrepareBackwardPage(int i) {
    Themis_DebugAssert(GetCache(i).state == eCacheStateEmpty);
    int nNext = InnerGetRotatedIndex(i + 1);
    Themis_DebugAssert(GetCache(nNext).state != eCacheStateLoading);
    PrepareBackwardPage(*GetCache(i).pageData, *GetCache(nNext).pageData);
}

/// ݂̂ǂݍ
void CPageLoaderBase::InnerLoadCurrent() {
    Themis_DebugAssert(GetCache(m_nCurrent).state == eCacheStateEmpty);
    // ǂݍ
    std::tr1::shared_ptr<IPageData> pPage = GetCache(m_nCurrent).pageData;
    std::tr1::shared_ptr<ILoadData> pLoadData = PrepareLoadPage(*pPage);
    if (pLoadData != NULL && LoadPage(*pLoadData)) {
        DisposeLoadPage(*pPage, *pLoadData);
        GetCache(m_nCurrent).state = eCacheStateLoaded;
    } else {
        GetCache(m_nCurrent).state = eCacheStateLoadFailed;
    }
}

/// ǂ݃Xbh
void CPageLoaderBase::OnThread() {
    Thread::SetName(L"PageLoaderThread");
    while (threadValid) {
        std::tr1::shared_ptr<IPageData> page; // ǂݍރy[W
        std::tr1::shared_ptr<ILoadData> loadData;
        int cacheIndex;
        {
            boost::recursive_mutex::scoped_lock lock(mutex);
            if (!threadValid) return;
            // ǂݍނƂ낪oĂ܂ő҂[vB
            int neighborIndex; // ׂindex
            bool forward;
            while (!FindToLoad(cacheIndex, neighborIndex, forward)) {
                // ǂݍނƂȂ҂B
                cacheCondition.wait(lock);
                if (!threadValid) return;
            }
            // ǂݍݏB
            GetCache(cacheIndex).state = eCacheStateLoading; // ǂݍݒƂ
            page = GetCache(cacheIndex).pageData;
            if (forward) {
                PrepareForwardPage(*page, *GetCache(neighborIndex).pageData);
            } else {
                PrepareBackwardPage(*page, *GetCache(neighborIndex).pageData);
            }
            // hNX̓ǂݍݑO
            loadData = PrepareLoadPage(*page);
        }
        // ǂݍ݁B
        if (loadData == NULL || !LoadPage(*loadData)) {
            // ǂݍݎšnB
            boost::recursive_mutex::scoped_lock lock(mutex);
            if (!threadValid) return;
            OnEndLoad(cacheIndex, false);
        } else {
            // ǂݍݐ̌nB
            boost::recursive_mutex::scoped_lock lock(mutex);
            if (!threadValid) return;
            if (GetCache(cacheIndex).state == eCacheStateLoading) {
                // ǂݍ񂾂̂LbVɃZbgāAԂǂݍݍς݂ɁB
                DisposeLoadPage(*page, *loadData);
                OnEndLoad(cacheIndex, true);
            } else {
                Themis_DebugAssert(GetCache(cacheIndex).state == eCacheStateEmpty ||
                    (cacheIndex == m_nCurrent && GetCache(cacheIndex).state == eCacheStateLoaded));
                // LbVNAꂽۂ̂œǂݍ񂾃f[^͔jB
            }
        }
    }
}

/// ǂݍރLbVʒuTBȂfalse
bool CPageLoaderBase::FindToLoad(int& cacheIndex, int& neighborIndex, bool& forward) {
    if (!IsReady()) return false; // LbVoĂȂȂT܂łȂB
    for (int i = 0, n = GetForwardCount(); i < n; i++) {
        int cache = InnerGetRotatedIndex(m_nCurrent + i + 1);
        if (GetCache(cache).state == eCacheStateEmpty) {
            int neighbor = InnerGetRotatedIndex(cache - 1); // ׂ
            if (GetCache(neighbor).state == eCacheStateLoaded) { // ׂƓǂݍݍς݁B
                // ǂ܂ĂȂ̔!
                cacheIndex = cache;
                neighborIndex = neighbor;
                forward = true;
                return true;
            }
        }
    }
    for (int i = 0, n = GetBackwardCount(); i < n; i++) {
        int cache = InnerGetRotatedIndex(m_nCurrent - i - 1);
        if (GetCache(cache).state == eCacheStateEmpty) {
            int neighbor = InnerGetRotatedIndex(cache + 1); // ׂ
            if (GetCache(neighbor).state == eCacheStateLoaded) { // ׂƓǂݍݍς݁B
                // ǂ܂ĂȂ̔!
                cacheIndex = cache;
                neighborIndex = neighbor;
                forward = false;
                return true;
            }
        }
    }
    // ǂ܂ĂȂ̖ہB
    return false;
}

// ǂݍݏȈ
void CPageLoaderBase::OnEndLoad(int cacheIndex, bool succeed) {
    Themis_DebugAssert(GetCache(cacheIndex).state == eCacheStateLoading);
    GetCache(cacheIndex).state = succeed ? eCacheStateLoaded : eCacheStateLoadFailed;
    // ǂݍݑ҂Ă犮ʒm𑗂
    loadingCondition.notify_all();
}

}
