#include "stdafx.h"
#include "Sntp.h"
#include "Socket.h"
#include "Windows/NativeString.h"
#include "Windows/Process.h"
#include <winsock.h>
#include <time.h>
namespace Themis {

CSntp::CSntp() {
}

CSntp::~CSntp() {
}

int CSntp::SetClock(const string& strServer) {
    std::tr1::shared_ptr<Process> process = Process::GetCurrentProcess();
    ProcessPriorityClass::T priority = process->GetPriorityClass();
    process->SetPriorityClass(ProcessPriorityClass::RealTime);
    try {
        pair<int, int> p = GetLag(strServer);
        SetClock(p);
        process->SetPriorityClass(priority);
        return p.first * 1000 + p.second;
    } catch (...) {
        process->SetPriorityClass(priority);
        throw;
    }
}

namespace {
    inline int64_t ntohlLL(uint32_t dw) {
        return (int64_t)ntohl(dw);
    }
}
pair<int, int> CSntp::GetLag(const string& strServer) {
  #include <pshpack1.h>
    struct SNtpPacket {
        uint8_t nLVNByte;       // [bwq(2bits), o[Wԍ(3bits), [h(3bits)
        uint8_t byStratum;      // Kw
        int8_t nPollInterval;   // |[OԊu
        int8_t nPrecision;      // x
        uint32_t nRootDelay;    // [gx
        uint32_t nRootDisp;     // [ggU
        uint32_t nRefId;        // QƎʎq
        uint32_t nRefSec;       // QƎ
        uint32_t nRefFrac;      // QƎ
        uint32_t nOrgSec;       // Jn
        uint32_t nOrgFrac;      // Jn
        uint32_t nRcvSec;       // M
        uint32_t nRcvFrac;      // M
        uint32_t nTransSec;     // M
        uint32_t nTransFrac;    // M
    };
  #include <poppack.h>

    // \Pbg̏
    CUdpSocket oSocket;
    CUdpHost oServer(strServer, 123);

    // f[^̑MȂǂȂ
    SNtpPacket s = {0x23}, r;
    SYSTEMTIME st;
    ::GetSystemTime(&st);
    s.nTransSec = htonl(static_cast<uint32_t>(time(NULL) + 2208988800)); // 32rbgł񂩂ˁH
    s.nTransFrac = htonl(st.wMilliseconds * 4294967);
    oSocket.SendTo(oServer, &s, sizeof(s));
    if (!oSocket.Select(5)) {
        throw RuntimeException(L"T[o[̃f[^MɃ^CAEg܂");
    }
    oSocket.ReceiveFrom(oServer, &r, sizeof(r));
    ::GetSystemTime(&st);
    int64_t nRSec = (int64_t)time(NULL) + 0x83aa7e80u;
    int64_t nRFrac = Int32x32To64(st.wMilliseconds, 4294967);

    // f[^̃`FbNƂ덷vZƂ
    Themis_DebugAssert(s.nTransFrac == r.nOrgFrac);
    Themis_DebugAssert(s.nTransSec == r.nOrgSec);
    if ((r.nLVNByte & 0xc0) == 0xc0) {  // 0xc0 == (1 << 7) | (1 << 6)
        throw RuntimeException(L"SNTPŕsȃG[܂");
    }
    int64_t nFrac64 = (ntohlLL(r.nRcvFrac) - ntohlLL(r.nOrgFrac) + ntohlLL(r.nTransFrac) - nRFrac) / 2;
    int64_t nSec64 = ntohlLL(r.nRcvSec) - ntohlLL(r.nOrgSec) + ntohlLL(r.nTransSec) - nRSec;
    if (nSec64 % 2) {
        nFrac64 += 0x80000000; // 2^310.5b
    }
    nSec64 /= 2;
    nSec64 += (int)(nFrac64 / 0x100000000); // 2^32
    nFrac64 /= 4294967; // 2^32/1000
    Themis_DebugAssert(-1000 < (int)nFrac64 && (int)nFrac64 < 1000);
    return make_pair((int)nSec64, (int)nFrac64);
}

void CSntp::SetClock(const pair<int, int>& p) {
    HANDLE hToken = NULL;
    TOKEN_PRIVILEGES oTp = {0};
    if (NativeString::IsUnicode()) {
        ::OpenProcessToken(::GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken);
        ::LookupPrivilegeValue(NULL, SE_SYSTEMTIME_NAME, &oTp.Privileges[0].Luid);
        oTp.PrivilegeCount = 1;
        oTp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
        ::AdjustTokenPrivileges(hToken, FALSE, &oTp, 0, NULL, NULL);
    }

    time_t nCurrent = time(NULL);
    SYSTEMTIME st;
    ::GetSystemTime(&st);
    int nMilliSeconds = st.wMilliseconds + p.second;
    if (nMilliSeconds < 0) {
        nMilliSeconds += 1000;
        nCurrent--;
    } else if (1000 <= nMilliSeconds) {
        nMilliSeconds -= 1000;
        nCurrent++;
    }
    st.wMilliseconds = (WORD)nMilliSeconds;
    nCurrent += p.first;
    tm lt = *localtime(&nCurrent);
    st.wYear = lt.tm_year + 1900;
    st.wMonth = lt.tm_mon + 1;
    st.wDay = lt.tm_mday;
    st.wHour = lt.tm_hour;
    st.wMinute = lt.tm_min;
    st.wSecond = lt.tm_sec;
    ::SetLocalTime(&st);

    if (hToken != NULL) {
        oTp.Privileges[0].Attributes = 0;
        ::AdjustTokenPrivileges(hToken, FALSE, &oTp, 0, NULL, NULL);
        ::CloseHandle(hToken);
    }
}

}
