#pragma once
#include "Char.h"
#include "List.h"
#include <boost/tr1/memory.hpp>
#include <boost/operators.hpp>
#include <boost/assert.hpp>
#include <string>
#include <algorithm>
namespace Themis {

/// ̔r@
namespace StringComparison {
    enum T {
        CurrentCulture,
        CurrentCultureIgnoreCase,
        InvariantCulture,
        InvariantCultureIgnoreCase,
        Ordinal,
        OrdinalIgnoreCase
    };
}

/// Split̃IvV
namespace StringSplitOptions {
    enum T { None, RemoveEmptyEntries };
}

namespace NormalizationForm {
    enum T {
        /// SWgp Unicode 𐳋KA\ȏꍇ́AV[PXvȕŒu邱Ƃ܂B
        FormC = 1,
        /// SWgp Unicode 𐳋K邱Ƃ܂B
        FormD = 2,
        /// S݊gp Unicode 𐳋KA\ȏꍇ́AV[PXvȕŒu邱Ƃ܂B
        FormKC = 5,
        /// S݊gp Unicode 𐳋K邱Ƃ܂B
        FormKD = 6,
    };
}

namespace Detail {
    struct ToLowerFunctor          { inline wchar_t operator()(wchar_t c) { return Char::ToLower(c); } };
    struct ToLowerInvariantFunctor { inline wchar_t operator()(wchar_t c) { return Char::ToLowerInvariant(c); } };
    struct ToUpperFunctor          { inline wchar_t operator()(wchar_t c) { return Char::ToUpper(c); } };
    struct ToUpperInvariantFunctor { inline wchar_t operator()(wchar_t c) { return Char::ToUpperInvariant(c); } };
}

/**
 * .NET̓NXIȂ́B
 * rȂǂ́A悭̂őSwscsmpƂCB
 * Ordinal, OrdinalIgnoreCaseH
 */
class String : private boost::totally_ordered<String> {
    // addable͑Ώ̂ȂƎgȂBB
#ifdef NDEBUG
    inline void SetDebug() {}
#else
    std::wstring dbgStr;
    void SetDebug() { dbgStr = *p_; }
#endif
    std::tr1::shared_ptr<const std::wstring> p_;
protected:
    /// C#typedef(p)
    typedef String string;

public:
    /// ̕
    static const string Empty;

    /// 
    String() : p_(Empty.GetPtr()) { SetDebug(); }
    /// 
    String(const wchar_t* value) : p_(new std::wstring(value)) { SetDebug(); }
    /// 
    String(const std::wstring& value) : p_(new std::wstring(value)) { SetDebug(); }
    /// 
    String(const std::vector<wchar_t>& value) : p_(new std::wstring(value.begin(), value.end())) { SetDebug(); }
    /// ccountׂB
    String(wchar_t c, int count) : p_(new std::wstring(count, c)) { SetDebug(); }
    /// startIndexlength
    String(const wchar_t* value, int startIndex, int length)
        : p_(new std::wstring(value + startIndex, value + startIndex + length)) { SetDebug(); }
  #ifndef NDEBUG
    ~String() {
        // ɈAςȎistr̓eĎB
        BOOST_ASSERT(ToCharArray() == dbgStr);
    }
  #endif

    //@{ rZq
    bool operator==(const string& other) const {
        return GetPtr() == other.GetPtr() || ToCharArray() == other.ToCharArray();
    }
    bool operator<(const string& other) const {
        return ToCharArray() < other.ToCharArray();
    }
    //@}

    /// ǉ
    string& operator+=(const string& other) {
        SetPtr(new std::wstring(ToCharArray() + other.ToCharArray()));
        return *this;
    }
    string& operator+=(const wchar_t* other) {
        SetPtr(new std::wstring(ToCharArray() + other));
        return *this;
    }
    //@{ A
    string operator+(const string& other) const {
        return new std::wstring(ToCharArray() + other.ToCharArray());
    }
    string operator+(const wchar_t* other) const {
        return new std::wstring(ToCharArray() + other);
    }
    //@}

    /// ̎擾
    int GetLength() const { return static_cast<int>(ToCharArray().length()); }

    /// w肵ʒu̕擾
    wchar_t operator[](int n) const { return ToCharArray()[n]; }

    /// CX^X̕擾
  //string Clone() const { return *this; }

    //@{ r
    static int Compare(const string& strA, const string& strB) {
        return Compare(strA, strB, false);
    }
    static int Compare(const string& strA, const string& strB, bool ignoreCase) {
        if (strA.GetPtr() == strB.GetPtr()) {
            return 0; // ZRĂ݂
        }
        if (ignoreCase) {
            return _wcsicmp(strA.ToCharPtr(), strB.ToCharPtr());
        } else {
            return wcscmp(strA.ToCharPtr(), strB.ToCharPtr());
        }
    }
    static int Compare(const string& strA, int indexA, const string& strB, int indexB, int length, StringComparison::T comparisonType) {
        BOOST_ASSERT(0 <= indexA && 0 <= indexB && 0 <= length);
        switch (comparisonType) {
        default:
        case StringComparison::CurrentCulture:
        case StringComparison::InvariantCulture:
        case StringComparison::Ordinal:
            return wcsncmp(strA.ToCharPtr() + indexA, strB.ToCharPtr() + indexB, length);
        case StringComparison::CurrentCultureIgnoreCase:
        case StringComparison::InvariantCultureIgnoreCase:
        case StringComparison::OrdinalIgnoreCase:
            return _wcsnicmp(strA.ToCharPtr() + indexA, strB.ToCharPtr() + indexB, length);
        }
    }
    //@}
    //@{ lIȔrH
    static int CompareOrdinal(const string& strA, const string& strB) {
        return Compare(strA, strB);
    }
    static int CompareOrdinal(const string& strA, int indexA, const string& strB, int indexB, int length) {
        return Compare(strA, indexA, strB, indexB, length, StringComparison::Ordinal);
    }
    //@}
    /// r
    int CompareTo(const string& strB) const { return Compare(*this, strB); }

    /// A
    static string Concat(const std::vector<string>& values) {
        std::auto_ptr<std::wstring> p(new std::wstring());
        for (size_t i = 0; i < values.size(); i++) {
            *p += values[i].ToCharArray();
        }
        return p.release();
    }

    /// value܂łtrue
    bool Contains(const string& value) const {
        return ToCharArray().find(value.ToCharArray()) != std::wstring::npos;
    }

    /// Rs[쐬
    //static string Copy(const string& str) { return str.Clone(); }

    /// Rs[B
    void CopyTo(int sourceIndex, wchar_t* destination, int destinationIndex, int count) const {
        BOOST_ASSERT(0 <= sourceIndex && 0 <= count && sourceIndex + count <= GetLength());
        memcpy(destination + destinationIndex, ToCharPtr() + sourceIndex, sizeof(wchar_t) * count);
    }

    //@{ ǂ
    static bool Equals(const string& a, const string& b) {
        return a == b;
    }
    static bool Equals(const string& a, const string& b, StringComparison::T comparisonType) {
        switch (comparisonType) {
        default:
        case StringComparison::CurrentCulture:
        case StringComparison::InvariantCulture:
        case StringComparison::Ordinal:
            return Equals(a, b);
        case StringComparison::CurrentCultureIgnoreCase:
        case StringComparison::InvariantCultureIgnoreCase:
        case StringComparison::OrdinalIgnoreCase:
            return Compare(a, b, true) == 0;
        }
    }
    bool Equals(const string& value) const {
        return Equals(*this, value);
    }
    bool Equals(const string& value, StringComparison::T comparisonType) const {
        return Equals(*this, value, comparisonType);
    }
    //@}

    /// FormatBC#݊ł͂Ȃprintf`ŁB
    static string Format(const wchar_t* format, ...);

    /// nbVR[h̎Zo
    //int GetHashCode() const { BOOST_ASSERT(0 && ""); }

    //@{ ŏɌʒuԂ
    int IndexOf(wchar_t value) const {
        return IndexOf(value, 0);
    }
    int IndexOf(wchar_t value, int startIndex) const {
        return IndexOf(value, startIndex, GetLength() - startIndex);
    }
    int IndexOf(wchar_t value, int startIndex, int count) const {
        BOOST_ASSERT(0 <= startIndex && 0 <= count && startIndex + count <= GetLength());
        for (int i = 0; i < count; i++) {
            if (ToCharArray()[startIndex + i] == value) {
                return startIndex + i; // I
            }
        }
        return -1;
    }
    int IndexOf(const string& value) const {
        return IndexOf(value, 0);
    }
    int IndexOf(const string& value, int startIndex) const {
        return IndexOf(value, startIndex, GetLength() - startIndex);
    }
    int IndexOf(const string& value, int startIndex, int count) const {
        return IndexOf(value, startIndex, count, StringComparison::CurrentCulture);
    }
    int IndexOf(const string& value, StringComparison::T comparisonType) const {
        return IndexOf(value, 0, comparisonType);
    }
    int IndexOf(const string& value, int startIndex, StringComparison::T comparisonType) const {
        return IndexOf(value, startIndex, GetLength() - startIndex, comparisonType);
    }
    int IndexOf(const string& value, int startIndex, int count, StringComparison::T comparisonType) const {
        BOOST_ASSERT(0 <= startIndex && 0 <= count && startIndex + count <= GetLength());
        for (int i = 0; i < count; i++) {
            if (Compare(*this, startIndex + i, value, 0, value.GetLength(), comparisonType) == 0) {
                return startIndex + i; // I
            }
        }
        return -1;
    }
    int IndexOfAny(const std::vector<wchar_t>& anyOf) const {
        return IndexOfAny(anyOf, 0);
    }
    int IndexOfAny(const std::vector<wchar_t>& anyOf, int startIndex) const {
        return IndexOfAny(anyOf, startIndex, GetLength() - startIndex);
    }
    int IndexOfAny(const std::vector<wchar_t>& anyOf, int startIndex, int count) const {
        BOOST_ASSERT(0 <= startIndex && 0 <= count && startIndex + count <= GetLength());
        for (int i = 0; i < count; i++) {
            if (std::find(anyOf.begin(), anyOf.end(), ToCharArray()[startIndex + i]) != anyOf.end()) {
                return startIndex + i; // I
            }
        }
        return -1;
    }
    //@}

    //@{ Ōɏoʒu擾
    int LastIndexOf(wchar_t value) const {
        return LastIndexOf(value, 0);
    }
    int LastIndexOf(wchar_t value, int startIndex) const {
        return LastIndexOf(value, startIndex, GetLength() - startIndex);
    }
    int LastIndexOf(wchar_t value, int startIndex, int count) const {
        BOOST_ASSERT(0 <= startIndex && 0 <= count && startIndex + count <= GetLength());
        for (int i = count - 1; 0 <= i; i--) { // IndexOfƂ͂Ⴄ
            if (ToCharArray()[startIndex + i] == value) {
                return startIndex + i; // I
            }
        }
        return -1;
    }
    int LastIndexOf(const string& value) const {
        return LastIndexOf(value, 0);
    }
    int LastIndexOf(const string& value, int startIndex) const {
        return LastIndexOf(value, startIndex, GetLength() - startIndex);
    }
    int LastIndexOf(const string& value, int startIndex, int count) const {
        return LastIndexOf(value, startIndex, count, StringComparison::CurrentCulture);
    }
    int LastIndexOf(const string& value, StringComparison::T comparisonType) const {
        return LastIndexOf(value, 0, comparisonType);
    }
    int LastIndexOf(const string& value, int startIndex, StringComparison::T comparisonType) const {
        return LastIndexOf(value, startIndex, GetLength() - startIndex, comparisonType);
    }
    int LastIndexOf(const string& value, int startIndex, int count, StringComparison::T comparisonType) const {
        BOOST_ASSERT(0 <= startIndex && 0 <= count && startIndex + count <= GetLength());
        for (int i = count - 1; 0 <= i; i--) { // IndexOfƂ͂Ⴄ
            if (Compare(*this, startIndex + i, value, 0, value.GetLength(), comparisonType) == 0) {
                return startIndex + i; // I
            }
        }
        return -1;
    }
    int LastIndexOfAny(const std::vector<wchar_t>& anyOf) const {
        return LastIndexOf(anyOf, 0);
    }
    int LastIndexOfAny(const std::vector<wchar_t>& anyOf, int startIndex) const {
        return LastIndexOf(anyOf, startIndex, GetLength() - startIndex);
    }
    int LastIndexOfAny(const std::vector<wchar_t>& anyOf, int startIndex, int count) const {
        BOOST_ASSERT(0 <= startIndex && 0 <= count && startIndex + count <= GetLength());
        for (int i = count - 1; 0 <= i; i--) { // IndexOfƂ͂Ⴄ
            if (std::find(anyOf.begin(), anyOf.end(), ToCharArray()[startIndex + i]) != anyOf.end()) {
                return startIndex + i; // I
            }
        }
        return -1;
    }
    //@}

    /// }stringԂB
    string Insert(int startIndex, const string& value) const {
        std::auto_ptr<std::wstring> p(new std::wstring(ToCharArray()));
        p->insert(startIndex, value.ToCharArray());
        return p.release();
    }

    /// w肵 System.String ւ̃VXeQƂ擾܂B
    //static string Intern(const string& str);
    /// w肵 System.String ւ̃VXeQƂ擾܂B
    //static string IsInterned(const string& str);

    //@{ UnicodeKsԂ܂B
    string Normalize() const {
        return Normalize(NormalizationForm::FormC);
    }
    string Normalize(NormalizationForm::T normalizationForm) const;
    //@}

    //@{ ̕UnicodeK`ł邩ǂ܂B
    bool IsNormalized() const {
        return IsNormalized(NormalizationForm::FormC);
    }
    bool IsNormalized(NormalizationForm::T normalizationForm) const;
    //@}

    //@{ Ȃtrue
    static bool IsNullOrEmpty(const string& value) { return value.ToCharArray().empty(); }
    static bool IsNullOrEmpty(const string* value) { return value == NULL || value->ToCharArray().empty(); }
    static bool IsNullOrEmpty(const std::auto_ptr<string>& value) { return value.get() == NULL || value->ToCharArray().empty(); }
    static bool IsNullOrEmpty(const std::tr1::shared_ptr<string>& value) { return value.get() == NULL || value->ToCharArray().empty(); }
    //@}

    //@{ ̃CX^X̕E񂹂Aw肵̕ɂȂ܂ŁAɋ󔒂𖄂ߍ݂܂B
    string PadLeft(int totalWidth) const {
        return PadLeft(totalWidth, L' ');
    }
    string PadLeft(int totalWidth, wchar_t paddingChar) const {
        if (totalWidth <= GetLength()) {
            return *this;
        }
        return Insert(0, string(paddingChar, GetLength() - totalWidth));
    }
    //@}

    //@{ ̕̕񂹂Aw肳ꂽ̒ɂȂ܂ŁAEɋ󔒂𖄂ߍ݂܂B
    string PadRight(int totalWidth) const {
        return PadRight(totalWidth, L' ');
    }
    string PadRight(int totalWidth, wchar_t paddingChar) const {
        if (totalWidth <= GetLength()) {
            return *this;
        }
        std::auto_ptr<std::wstring> p(new std::wstring(ToCharArray()));
        for (int i = 0, n = GetLength() - totalWidth; i < n; i++) {
            *p += paddingChar;
        }
        return p.release();
    }
    //@}

    //@{ w肳ꂽʒu̕̕񂩂폜
    string Remove(int startIndex) const {
        return Remove(startIndex, 1);
    }
    string Remove(int startIndex, int count) const {
        BOOST_ASSERT(0 <= startIndex && 0 <= count && startIndex + count <= GetLength());
        if (count <= 0) {
            return *this;
        }
        std::auto_ptr<std::wstring> p(new std::wstring());
        p->reserve(GetLength() - count);
        if (0 < startIndex) {
            p->append(ToCharPtr(), startIndex);
        }
        p->append(ToCharPtr() + startIndex + count,  GetLength() - count - startIndex);
        return p.release();
    }
    //@}
    //@{ ̃CX^Xɏow蕶SāAʂɎw肵ɒu
    string Replace(wchar_t oldChar, wchar_t newChar) const {
        std::auto_ptr<std::wstring> p(new std::wstring(ToCharArray()));
        std::replace(p->begin(), p->end(), oldChar, newChar);
        return p.release();
    }
    string Replace(const string& oldValue, const string& newValue) const;
    //@}

    //@{ ̃CX^X́Aw肳ꂽŋ؂ꂽi[vector<string>Ԃ
    List<string> Split(const std::vector<wchar_t>& separator) const {
        return Split(separator, INT_MAX);
    }
    List<string> Split(const std::vector<wchar_t>& separator, int count) const {
        return Split(separator, count, StringSplitOptions::None);
    }
    List<string> Split(const std::vector<wchar_t>& separator, StringSplitOptions::T options) const {
        return Split(separator, INT_MAX, options);
    }
    List<string> Split(const std::vector<wchar_t>& separator, int count, StringSplitOptions::T options) const;
    List<string> Split(const std::vector<string>& separator) const {
        return Split(separator, INT_MAX, StringSplitOptions::None);
    }
    List<string> Split(const std::vector<string>& separator, int count, StringSplitOptions::T options) const;
    //@}

    //@{ separatorŘA
    static string Join(const string& separator, const std::vector<string>& value) {
        return Join(separator, value, 0, static_cast<int>(value.size()));
    }
    static string Join(const string& separator, const std::vector<string>& value, int startIndex, int count);
    //@}

    //@{ 擪v邩ǂ𔻒
    bool StartsWith(const string& value) const {
        return StartsWith(value, false);
    }
    bool StartsWith(const string& value, bool ignoreCase) const {
        return StartsWith(value, ignoreCase ?
            StringComparison::CurrentCultureIgnoreCase :
            StringComparison::CurrentCulture);
    }
    bool StartsWith(const string& value, StringComparison::T comparisonType) const {
        return value.GetLength() <= GetLength() &&
            Compare(*this, 0, value, 0, value.GetLength(), comparisonType) == 0;
    }
    //@}
    //@{ I[v邩ǂ
    bool EndsWith(const string& value) const {
        return EndsWith(value, false);
    }
    bool EndsWith(const string& value, bool ignoreCase) const {
        return EndsWith(value, ignoreCase ?
            StringComparison::CurrentCultureIgnoreCase :
            StringComparison::CurrentCulture);
    }
    bool EndsWith(const string& value, StringComparison::T comparisonType) const {
        return value.GetLength() <= GetLength() &&
            Compare(*this, GetLength() - value.GetLength(),
                value, 0, value.GetLength(), comparisonType) == 0;
    }
    //@}

    //@{ ̎擾
    string Substring(int startIndex) const {
        return Substring(startIndex, GetLength() - startIndex);
    }
    string Substring(int startIndex, int length) const {
        BOOST_ASSERT(0 <= startIndex && 0 <= length && startIndex + length <= GetLength());
        if (length <= 0) {
            return Empty;
        } else if (GetLength() <= length) {
            return *this; // ܂薳P[XȂǂĐߖ񂵂Ă݂
        }
        std::wstring::const_iterator it = ToCharArray().begin() + startIndex;
        return new std::wstring(it, it + length);
    }
    //@}

    //@{ charz
    const std::wstring& ToCharArray() const {
        return *GetPtr();
    }
    std::wstring ToCharArray(int startIndex, int count) const {
        BOOST_ASSERT(0 <= startIndex && 0 <= count && startIndex + count <= GetLength());
        return ToCharArray().substr(startIndex, count);
    }
    //@}

    //@{ 
    string ToLower() const {
        std::auto_ptr<std::wstring> p(new std::wstring(GetLength(), L'\0'));
        std::transform(ToCharArray().begin(), ToCharArray().end(),
            p->begin(), Detail::ToLowerFunctor());
        return p.release();
    }
    string ToLowerInvariant() const {
        std::auto_ptr<std::wstring> p(new std::wstring(GetLength(), L'\0'));
        std::transform(ToCharArray().begin(), ToCharArray().end(),
            p->begin(), Detail::ToLowerInvariantFunctor());
        return p.release();
    }
    //@}
    //@{ 啶
    string ToUpper() const {
        std::auto_ptr<std::wstring> p(new std::wstring(GetLength(), L'\0'));
        std::transform(ToCharArray().begin(), ToCharArray().end(),
            p->begin(), Detail::ToUpperInvariantFunctor());
        return p.release();
    }
    string ToUpperInvariant() const {
        std::auto_ptr<std::wstring> p(new std::wstring(GetLength(), L'\0'));
        std::transform(ToCharArray().begin(), ToCharArray().end(),
            p->begin(), Detail::ToUpperInvariantFunctor());
        return p.release();
    }
    //@}

    //@{ Ő(܂͎w蕶)̏
    string Trim() const;
    string Trim(const std::vector<wchar_t>& trimChars) const;
    string TrimStart() const;
    string TrimStart(const std::vector<wchar_t>& trimChars) const;
    string TrimEnd() const;
    string TrimEnd(const std::vector<wchar_t>& trimChars) const;
    //@}

    /// obt@̎擾
    const wchar_t* ToCharPtr() const { return ToCharArray().c_str(); }

protected:
    //@{ |C^珉Bȍ~̃|C^oRł̏͋֎~B
    String(std::wstring* str) : p_(str) { SetDebug(); }
    String(const std::tr1::shared_ptr<const std::wstring>& str) : p_(str) { SetDebug(); }
    //@}

    //@{ |C^
    const std::tr1::shared_ptr<const std::wstring>& GetPtr() const { return p_; }
    void SetPtr(std::wstring* s) {
        p_.reset(s);
        SetDebug();
    }
    void SetPtr(const std::tr1::shared_ptr<const std::wstring>& s) {
        p_ = s;
        SetDebug();
    }
    //@}
};

/// C#L[[htypedef
typedef String string;

//@{ Z(A)Zq
inline string operator+(const wchar_t* strA, const string& strB) {
    return strA + strB.ToCharArray();
}
//@}

}
